from typing import (
    TYPE_CHECKING,
    Optional
)
from web3.contract import (
    ContractConstructor
)

from cfx_utils.decorators import (
    combomethod
)
from conflux_web3.types.transaction_hash import (
    TransactionHash
)
from conflux_web3.types import (
    TxParam,
    EpochNumberParam,
    EstimateResult
)
from conflux_web3._utils.transactions import (
    fill_transaction_defaults
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

class ConfluxContractConstructor(ContractConstructor):
    w3: "Web3"
    
    @combomethod
    def transact(self, transaction: Optional[TxParam] = None) -> TransactionHash:
        return super().transact(transaction)
    
    @combomethod
    def build_transaction(self, transaction: Optional[TxParam] = None) -> TxParam:
        built_transaction = self._build_transaction(transaction)
        return fill_transaction_defaults(self.w3, built_transaction)
    
    @combomethod
    def estimate_gas(
        self,
        transaction: Optional[TxParam] = None,
        block_identifier: Optional[EpochNumberParam] = None,
    ) -> EstimateResult:
        return self.estimate_gas_and_collateral(transaction, block_identifier)
    
    @combomethod
    def estimate_gas_and_collateral(
        self,
        transaction: Optional[TxParam] = None,
        block_identifier: Optional[EpochNumberParam] = None,
    ) -> EstimateResult:
        return super().estimate_gas(transaction, block_identifier)



from typing import (
    Union,
    overload,
)

import rlp
from hexbytes import HexBytes
from eth_utils import (
    keccak,
    to_bytes,
    to_checksum_address,
)

from cfx_address import Base32Address
from cfx_address.utils import (
    normalize_to
)
from cfx_utils.types import (
    ChecksumAddress,
)

def get_create_address(sender: Union[str, Base32Address], nonce: int, bytecode_hash: Union[bytes, str]) -> Union[ChecksumAddress, Base32Address]:
    """
    Determine the resulting `CREATE` opcode contract address for a sender and a nonce.
    Typically, the sender is a wallet address and the nonce is the next nonce of the sender.
    NOTE: in Conflux, the contract address is computed differently from that in Ethereum 
    where the bytecode hash is not accounted for in the address computation.
    
    :param sender: The address of the sender. Can be a hex string or a base32 address.
    :param nonce: The nonce of the sender.
    :param bytecode_hash: The keccak256 hash of the contract bytecode, whereas the "data" field of the transaction. Can be bytes or hex string.
    :return: The computed address as a hex string or base32 address, depending on the type of sender
    """
    address_hex = normalize_to(sender, None)
    contract_address = "0x8" + keccak(
        b"\x00"
        + to_bytes(hexstr=address_hex)
        + nonce.to_bytes(32, "little")
        + HexBytes(bytecode_hash)
    ).hex()[-39:]
    if Base32Address.is_valid_base32(sender):
        return Base32Address(contract_address, network_id=Base32Address(sender).network_id)
    return to_checksum_address(contract_address)


def get_create2_address(
    create2_factory_address: Union[str, Base32Address], salt: bytes, bytecode_hash: Union[bytes, str]
) -> Union[ChecksumAddress, Base32Address]:
    """
    Compute the address of a contract created using CREATE2.
    
    :param create2_factory_address: The address of the CREATE2 factory contract. On Conflux, it is deployed via `CIP-31 <https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-31.md>`_
    :param salt: A 32-byte value used as salt. Should be bytes.
    :param bytecode_hash: The keccak256 hash of the contract bytecode. Can be bytes or hex string.
    :return: The computed address as a hex string or base32 address, depending on the type of create2_factory_address
    """
    address_hex = normalize_to(create2_factory_address, None)
    contract_address = "0x8" + keccak(
        b"\xff"
        + to_bytes(hexstr=address_hex)
        + salt
        + HexBytes(bytecode_hash)
    ).hex()[-39:]
    if Base32Address.is_valid_base32(create2_factory_address):
        return Base32Address(contract_address, network_id=Base32Address(create2_factory_address).network_id)
    return to_checksum_address(contract_address)



import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)
from hexbytes import HexBytes

from eth_utils.functional import (
    to_tuple,
)
from web3.contract.base_contract import (
    BaseContractEvent,
    BaseContractEvents,
)
from web3.datastructures import (
    AttributeDict,
    MutableAttributeDict,
)
from web3.types import (
    ABI,
)
from web3.logs import (
    DISCARD,
    IGNORE,
    STRICT,
    WARN,
)
from web3._utils.events import (
    EventLogErrorFlags,
)
from web3._utils.filters import (
    construct_event_filter_params
)
from web3.exceptions import (
    InvalidEventABI,
    LogTopicError,
    MismatchedABI,
)

from cfx_address import (
    Base32Address
)
from cfx_utils.decorators import (
    combomethod
)
from conflux_web3.types import (
    AddressParam,
    TxReceipt,
    EventData,
    LogReceipt,
    EpochNumberParam,
)
from conflux_web3._utils.events import (
    cfx_get_event_data,
)
from conflux_web3._utils.decorators import (
    use_instead
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

class ConfluxContractEvent(BaseContractEvent):
    
    w3: "Web3"
    address: Base32Address
    
    @combomethod
    def process_receipt(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags = WARN
    ) -> Sequence[EventData]:
        return self._parse_logs(txn_receipt, errors) # type: ignore
    
    @combomethod
    def processReceipt(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags = WARN
    ) -> Sequence[EventData]:
        return self.process_receipt(txn_receipt, errors)
    
    @combomethod
    @to_tuple
    def _parse_logs(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags
    ) -> Iterable[EventData]:
        try:
            errors.name
        except AttributeError:
            raise AttributeError(
                f"Error flag must be one of: {EventLogErrorFlags.flag_options()}"
            )
        for transaction_log_index in range(len(txn_receipt["logs"])):
        # for log in txn_receipt["logs"]:
            log = txn_receipt["logs"][transaction_log_index]
            try:
                log = cast(LogReceipt, dict(log))
                log["transactionHash"] = txn_receipt["transactionHash"]
                log["blockHash"] = txn_receipt["blockHash"]
                log["epochNumber"] = txn_receipt["epochNumber"]
                log["transactionIndex"] = txn_receipt["index"]
                log["transactionLogIndex"] = transaction_log_index
                abi = self.abi or self._get_event_abi()
                rich_log = cfx_get_event_data(self.w3.codec, abi, log, self.w3.cfx.chain_id)
            except (MismatchedABI, LogTopicError, InvalidEventABI, TypeError) as e:
                if errors == DISCARD:
                    continue
                elif errors == IGNORE:
                    # type ignores b/c rich_log set on 1092 conflicts with mutated types
                    new_log = MutableAttributeDict(log)  # type: ignore
                    new_log["errors"] = e
                    rich_log = AttributeDict(new_log)  # type: ignore
                elif errors == STRICT:
                    raise e
                else:
                    warnings.warn(
                        f"The log with transaction hash: {txn_receipt['transactionHash']!r} "
                        f"encountered the following "
                        f"error during processing: {type(e).__name__}({e}). It has "
                        "been discarded."
                    )
                    continue
            yield rich_log # type: ignore

    @combomethod
    def process_log(self, log: LogReceipt) -> EventData:
        abi = self.abi or self._get_event_abi()
        return cfx_get_event_data(self.w3.codec, abi, log, self.w3.cfx.chain_id)

    @combomethod
    def processLog(self, log: LogReceipt) -> EventData:
        return self.process_log(self.w3.codec, self.abi, log)

    @combomethod
    @use_instead
    def createFilter(*args, **kwargs):
        pass
    
    @combomethod
    @use_instead
    def build_filter(*args, **kwargs):
        pass

    @combomethod
    def get_filter_topics(
        self,
        argument_filters: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Any:
        """
        Web3.py's createFilter interface is used for eth_newFilter RPCEndpoint.
        We disable the createFilter method and use get_filter_topics instead
        """
        if argument_filters is not None and len(kwargs.keys()) != 0:
            raise ValueError("Redundant Param: argument_filters is already provided")

        if argument_filters is None:
            argument_filters = kwargs

        _filters = dict(**argument_filters)

        event_abi = self._get_event_abi()

        BaseContractEvent.check_for_forbidden_api_filter_arguments(event_abi, _filters)

        _, event_filter_params = construct_event_filter_params(
            self._get_event_abi(),
            self.w3.codec,
            argument_filters=_filters,
        )
        return event_filter_params["topics"] # type: ignore
    
    @combomethod
    def get_logs(
        self,
        argument_filters: Optional[Dict[str, Any]] = None,
        fromEpoch: Optional[EpochNumberParam] = None,
        toEpoch: Optional[EpochNumberParam] = None,
        blockHashes: Optional[Sequence[HexBytes]] = None,
        address: Optional[Union[Base32Address, Sequence[Base32Address]]]=None
    ) -> Sequence[EventData]:
        """
        _summary_

        Parameters
        ----------
        argument_filters : Optional[Dict[str, Any]], optional
            _description_, by default None
        fromEpoch : Optional[EpochNumberParam], optional
            _description_, by default None
        toEpoch : Optional[EpochNumberParam], optional
            _description_, by default None
        blockHashes : Optional[Sequence[HexBytes]], optional
            _description_, by default None
        address : Optional[Union[Base32Address, Sequence[Base32Address]]], optional
            _description_, by default None

        Returns
        -------
        Iterable[EventData]
            _description_
        """        
        topics = self.get_filter_topics(argument_filters)
        filter_params = {
            "blockHashes": blockHashes,
            "fromEpoch": fromEpoch,
            "toEpoch": toEpoch,
            "topics": topics,
            "address": address
        }
        logs = self.w3.cfx.get_logs(filter_params)
        return tuple(
            self.process_log(log) for log in logs
        )
        
    @combomethod
    def getLogs(self, *args, **kwargs):
        return self.get_logs(*args, **kwargs)

class ConfluxContractEvents(BaseContractEvents):
    def __init__(
        self, abi: ABI, w3: "Web3", address: Optional[AddressParam] = None
    ) -> None:
        super().__init__(abi, w3, ConfluxContractEvent, address) # type: ignore
        
    def __getitem__(self, event_name: str) -> Type["ConfluxContractEvent"]:
        return cast(Type[ConfluxContractEvent], super().__getitem__(event_name))
    
    def __getattr__(self, event_name: str) -> Type["ConfluxContractEvent"]:
        return cast(Type[ConfluxContractEvent], super().__getattr__(event_name))



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

from eth_abi.codec import (
    ABICodec
)
from web3 import (
    Web3 as OriWeb3
)
from web3.providers.base import (
    BaseProvider,
)
from web3.types import (
    RPCEndpoint
)
from web3._utils.empty import (
    empty,
    Empty,
)
from web3.module import (
    Module,
)

from cfx_address import (
    Base32Address
)
from cfx_account import (
    Account
)
from conflux_web3.types import (
    MiddlewareOnion
)
from conflux_web3.middleware import (
    conflux_default_middlewares
)

from conflux_web3.client import (
    ConfluxClient
)
from conflux_web3.txpool import (
    Txpool
)
from conflux_web3.exceptions import (
    DeploymentInfoNotFound
)
from conflux_web3._utils.abi import (
    build_cfx_default_registry
)
from cns import (
    CNS
)
if TYPE_CHECKING:
    from conflux_web3.middleware.wallet import Wallet

# The module name __name__ should be Web3 
class Web3(OriWeb3):
    cfx: ConfluxClient
    txpool: Txpool
    
    def __init__(
        self,
        provider: BaseProvider,
        middlewares: Optional[Sequence[Any]] = None,
        modules: Optional[Dict[str, Union[Type[Module], Sequence[Any]]]] = None,
        # external_modules: Optional[Dict[str, Union[Type[Module], Sequence[Any]]]] = None,
        cns: CNS = cast(CNS, empty),
        **kwargs,
    ):
        """
        initialize a ``Web3`` instance to interact with the blockchain
        
        >>> w3 = Web3(Web3.HTTPProvider("http://127.0.0.1:12537"))
        >>> w3 = Web3(Web3.HTTPProvider("https://test.confluxrpc.com"))
        >>> w3 = Web3(Web3.HTTPProvider("https://main.confluxrpc.com"))

        Parameters
        ----------
        provider : BaseProvider
            ``provider`` decides on how to connect to the blockchain, 
        middlewares : Optional[Sequence[Any]], optional
            middlewares to use, recommended not to specify
        modules : Optional[Dict[str, Union[Type[Module], Sequence[Any]]]], optional
            modules to use, recommended not to specify
        cns : CNS, optional
            a cns object. If cns is not specified, w3.cns will be inited with default setting
        """        
        # ConfluxClient as eth provider, default middlewares as [] rather than None
        # OriWeb3.__init__(self, provider=provider, middlewares=middlewares, modules={
        #     "eth": ConfluxClient
        # })
        if middlewares is None:
            middlewares = conflux_default_middlewares(self)
        # TODO: finish provider default value logic
        self.manager = self.RequestManager(self, provider, middlewares)
        # this codec gets used in the module initialization,
        # so it needs to come before attach_modules
        self.codec = ABICodec(build_cfx_default_registry())

        if modules is None:
            # modules = get_default_modules()
            modules = {
                "cfx": ConfluxClient,
                "txpool": Txpool,
            }
        
        if "cfx" not in modules:
            raise ValueError("cfx module is missing in modules: cfx module is required to initialze a web3 instance")

        self.attach_modules(modules)  # type: ignore

        # if external_modules is not None:
        #     self.attach_modules(external_modules)

        
        # use __setattr__ to avoid language server type hint
        self.__setattr__("eth", self.cfx)
        
        # provide easy access
        # self.account = self.cfx.account
        # self.address = self.cfx.address
        
        # ens argument is remained for compatibility
        # but it is not allowed to specify ens AND cns at the same time

        # kwargs["ens"] might be None, we assume a developer will not deliberately pass empty
        if kwargs.get("ens", empty) is not empty:
            ens = kwargs["ens"]
            if cns is empty:
                cns = ens
            else:
                raise ValueError("Redundant arguments: ens and cns argument are both specified. Only ens argument OR cns argument is allowed")
        
        if cns == empty and self.is_connected():
            try:
                cns = CNS.from_web3(self)
            except DeploymentInfoNotFound:
                pass

        self.cns = cns
        
        # TODO: set contract
        
    @property
    def account(self) -> Account:
        return self.cfx.account
        
    @property
    def address(self) -> Type[Base32Address]:
        return self.cfx.address

    @property
    def api(self) -> str:
        from conflux_web3 import __version__
        return __version__

    @property
    def ens(self) -> CNS:
        return self.cns
    
    @property
    def cns(self) -> CNS:
        return self._ens # type: ignore
    
    @cns.setter
    def cns(self, new_cns: CNS) -> None:
        # TODO: check middlewares, modules and default account if cns is inited in this way
        self._ens = new_cns
        
    @ens.setter
    def ens(self, new_cns: CNS) -> None:
        self.cns = new_cns

    @property
    def middleware_onion(self) -> MiddlewareOnion:
        return cast(MiddlewareOnion, self.manager.middleware_onion)
    
    @property
    def clientVersion(self) -> str:
        return self.client_version
    
    @property
    def client_version(self) -> str:
        return self.cfx.client_version
    
    @property
    def wallet(self) -> "Wallet":
        return self.middleware_onion.get("wallet", None) # type: ignore
    
    def is_connected(self) -> bool:
        try:
            response = self.provider.make_request(RPCEndpoint("cfx_clientVersion"), [])
        except OSError:
            return False

        assert response["jsonrpc"] == "2.0" # type: ignore
        assert "error" not in response # type: ignore

        return True



# should be placed at the most front
import _web3_hook

from importlib.metadata import version

from conflux_web3.main import Web3
from conflux_web3.dev import (
    get_local_web3,
    get_mainnet_web3,
    get_testnet_web3
)
HTTPProvider = Web3.HTTPProvider

__version__ = version("conflux_web3")

__all__ = [
    "Web3",
    "HTTPProvider",
    "get_local_web3",
    "get_mainnet_web3",
    "get_testnet_web3",
]



from web3.types import (
    RPCEndpoint,
)

class RPC:
    # cfx
    cfx_getStatus = RPCEndpoint("cfx_getStatus")
    cfx_call = RPCEndpoint("cfx_call")
    cfx_checkBalanceAgainstTransaction = RPCEndpoint("cfx_checkBalanceAgainstTransaction")
    cfx_clientVersion = RPCEndpoint("cfx_clientVersion")
    cfx_epochNumber = RPCEndpoint("cfx_epochNumber")
    cfx_estimateGasAndCollateral = RPCEndpoint("cfx_estimateGasAndCollateral")
    cfx_gasPrice = RPCEndpoint("cfx_gasPrice")
    cfx_getAccount = RPCEndpoint("cfx_getAccount")
    cfx_getAccumulateInterestRate = RPCEndpoint("cfx_getAccumulateInterestRate")
    cfx_getAdmin = RPCEndpoint("cfx_getAdmin")
    cfx_getBalance = RPCEndpoint("cfx_getBalance")
    cfx_getBestBlockHash = RPCEndpoint("cfx_getBestBlockHash")
    cfx_getBlockByEpochNumber = RPCEndpoint("cfx_getBlockByEpochNumber")
    cfx_getBlockByBlockNumber = RPCEndpoint("cfx_getBlockByBlockNumber")
    cfx_getBlockByHash = RPCEndpoint("cfx_getBlockByHash")
    cfx_getBlockByHashWithPivotAssumption = RPCEndpoint("cfx_getBlockByHashWithPivotAssumption")
    cfx_getBlockRewardInfo = RPCEndpoint("cfx_getBlockRewardInfo")
    cfx_getBlocksByEpoch = RPCEndpoint("cfx_getBlocksByEpoch")
    cfx_getEpochReceipts = RPCEndpoint("cfx_getEpochReceipts")
    cfx_maxPriorityFeePerGas = RPCEndpoint("cfx_maxPriorityFeePerGas")
    cfx_feeHistory = RPCEndpoint("cfx_feeHistory")
    cfx_getFeeBurnt = RPCEndpoint("cfx_getFeeBurnt")

    cfx_getCode = RPCEndpoint("cfx_getCode")
    cfx_getCollateralForStorage = RPCEndpoint("cfx_getCollateralForStorage")
    cfx_getConfirmationRiskByHash = RPCEndpoint("cfx_getConfirmationRiskByHash")
    cfx_getDepositList = RPCEndpoint("cfx_getDepositList")
    cfx_getInterestRate = RPCEndpoint("cfx_getInterestRate")
    cfx_getLogs = RPCEndpoint("cfx_getLogs")
    cfx_getNextNonce = RPCEndpoint("cfx_getNextNonce")
    cfx_getSkippedBlocksByEpoch = RPCEndpoint("cfx_getSkippedBlocksByEpoch")
    cfx_getSponsorInfo = RPCEndpoint("cfx_getSponsorInfo")
    cfx_getStakingBalance = RPCEndpoint("cfx_getStakingBalance")
    cfx_getStorageAt = RPCEndpoint("cfx_getStorageAt")
    cfx_getStorageRoot = RPCEndpoint("cfx_getStorageRoot")
    cfx_getTransactionByHash = RPCEndpoint("cfx_getTransactionByHash")
    cfx_getTransactionReceipt = RPCEndpoint("cfx_getTransactionReceipt")
    cfx_getVoteList = RPCEndpoint("cfx_getVoteList")
    cfx_sendRawTransaction = RPCEndpoint("cfx_sendRawTransaction")
    cfx_sendTransaction = RPCEndpoint("cfx_sendTransaction")
    
    cfx_newFilter = RPCEndpoint("cfx_newFilter")
    cfx_newBlockFilter = RPCEndpoint("cfx_newBlockFilter")
    cfx_newPendingTransactionFilter = RPCEndpoint("cfx_newPendingTransactionFilter")
    cfx_getFilterChanges = RPCEndpoint("cfx_getFilterChanges")
    cfx_getFilterLogs = RPCEndpoint("cfx_getFilterLogs")
    cfx_uninstallFilter = RPCEndpoint("cfx_uninstallFilter")

    cfx_getPoSEconomics = RPCEndpoint("cfx_getPoSEconomics")
    cfx_getPoSRewardByEpoch = RPCEndpoint("cfx_getPoSRewardByEpoch")
    cfx_getParamsFromVote = RPCEndpoint("cfx_getParamsFromVote")
    cfx_getSupplyInfo = RPCEndpoint("cfx_getSupplyInfo")
    cfx_getCollateralInfo = RPCEndpoint("cfx_getCollateralInfo")

    cfx_getAccountPendingInfo = RPCEndpoint("cfx_getAccountPendingInfo")
    cfx_getAccountPendingTransactions = RPCEndpoint("cfx_getAccountPendingTransactions")
    cfx_checkBalanceAgainstTransaction = RPCEndpoint("cfx_checkBalanceAgainstTransaction")

    cfx_newFilter = RPCEndpoint("cfx_newFilter")
    cfx_newBlockFilter = RPCEndpoint("cfx_newBlockFilter")
    cfx_newPendingTransactionFilter = RPCEndpoint("cfx_newPendingTransactionFilter")
    cfx_getFilterChanges = RPCEndpoint("cfx_getFilterChanges")
    cfx_getFilterLogs = RPCEndpoint("cfx_getFilterLogs")
    cfx_uninstallFilter = RPCEndpoint("cfx_uninstallFilter")

    # only available in LocalRPC
    cfx_getTransactionsByEpoch = RPCEndpoint("cfx_getTransactionsByEpoch")
    cfx_getTransactionsByBlock = RPCEndpoint("cfx_getTransactionsByBlock")
    
    # trace
    trace_block = RPCEndpoint("trace_block")
    trace_transaction = RPCEndpoint("trace_transaction")
    trace_epoch = RPCEndpoint("trace_epoch")

    # debug
    accounts = RPCEndpoint("accounts")
    
    # txpool
    txpool_nextNonce = RPCEndpoint("txpool_nextNonce")

    # other
    # cfx_method = RPCEndpoint("cfx_method")

TRANSACTION_PARAMS_ABIS = {
    "from": "address",
    "to": "address",
    "data": "bytes",
    "gas": "uint",
    "gasPrice": "uint",
    "nonce": "uint",
    "value": "uint",
    "chainId": "uint",
    "storageLimit": "uint",
    "epochHeight": "uint",
}

# RPC request formatters, parses abi types 
# the complex ones (None parameter, such as EpochNumberParam) are implemented in _utils.method_formatters.py

EPOCH_NUMBER_PARAM = None

RPC_ABIS = {
    RPC.cfx_call: TRANSACTION_PARAMS_ABIS,
    RPC.cfx_estimateGasAndCollateral: TRANSACTION_PARAMS_ABIS,
    RPC.cfx_getBlockByHash: ["bytes32", "bool"],
    RPC.cfx_getBlockByEpochNumber: [EPOCH_NUMBER_PARAM, "bool"],
    RPC.cfx_getBlockByBlockNumber: [EPOCH_NUMBER_PARAM, "bool"],
    RPC.cfx_getBestBlockHash: [],
    RPC.cfx_getBlocksByEpoch: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getSkippedBlocksByEpoch: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getBlockByHashWithPivotAssumption: ["bytes32", "bytes32", "uint"],
    RPC.cfx_getEpochReceipts: [EPOCH_NUMBER_PARAM, "bool"],

    RPC.cfx_getBalance: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getStakingBalance: ["address", EPOCH_NUMBER_PARAM],
    
    RPC.cfx_getCode: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getStorageAt: ["address", "uint", EPOCH_NUMBER_PARAM],
    RPC.cfx_getStorageRoot: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getCollateralForStorage: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getAdmin: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getSponsorInfo: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getAccount: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getDepositList: ["address", EPOCH_NUMBER_PARAM],
    RPC.cfx_getVoteList: ["address", EPOCH_NUMBER_PARAM],
    
    RPC.cfx_getInterestRate: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getAccumulateInterestRate: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getBlockRewardInfo: [None], # integer epoch number, or the string "latest_checkpoint"
    RPC.cfx_getPoSEconomics: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getPoSRewardByEpoch: ["uint"],
    RPC.cfx_getParamsFromVote: [EPOCH_NUMBER_PARAM],
    RPC.cfx_getSupplyInfo: [],
    
    RPC.cfx_getAccountPendingInfo: ["addresss"],
    RPC.cfx_getAccountPendingTransactions: ["address", "uint", "uint"],
    RPC.cfx_checkBalanceAgainstTransaction: ["address", "address", "uint", "uint", "uint", EPOCH_NUMBER_PARAM],
    
    RPC.cfx_getTransactionByHash: ["bytes32"],
    RPC.cfx_getTransactionReceipt: ["bytes32"],
    RPC.cfx_getConfirmationRiskByHash: ["bytes32"],
    # "cfx_sendRawTransaction": ["bytes"],
    RPC.cfx_sendRawTransaction: ["bytes"],
    # "cfx_sendTransaction": TRANSACTION_PARAMS_ABIS,
    RPC.cfx_sendTransaction: TRANSACTION_PARAMS_ABIS,

    # "cfx_getLogs": FILTER_PARAMS_ABIS,
    # "cfx_signTransaction": TRANSACTION_PARAMS_ABIS,
    # "cfx_sign": ["address", "bytes"],
}



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    Dict,
    cast,
    overload
)
from typing_extensions import (
    Literal,
    Unpack,
)
import warnings
from hexbytes import HexBytes

from cytoolz import ( keyfilter, merge ) # type:ignore
from eth_typing.encoding import (
    HexStr
)
from eth_utils.toolz import (
    assoc  # type: ignore
)

from web3.eth import (
    BaseEth, 
    Eth
)
# empty means a empty address but is not None
from web3._utils.empty import (
    Empty,
    empty,
)
from web3._utils.threads import (
    Timeout,
)
from web3.types import (
    RPCEndpoint,
)
from web3.exceptions import (
    TransactionNotFound,
    TimeExhausted
)
from web3._utils.blocks import is_hex_encoded_block_hash as is_hash32_str

from cfx_utils.token_unit import (
    to_int_if_drip_units,
    AbstractDerivedTokenUnit
)
from cfx_utils.types import (
    _Hash32,
    Hash32,
    Drip,
    EpochLiteral,
    EpochNumberParam,
    AddressParam,
    TxParam,
    EpochNumber,
    Storage,
)
from cfx_address import (
    Base32Address,
)
from cfx_address.address import (
    get_base32_address_factory,
)
from cfx_account import (
    Account,
    LocalAccount,
)

from conflux_web3._utils.decorators import (
    cached_property
)
from conflux_web3._utils.rpc_abi import (
    RPC
)
from conflux_web3._utils.disabled_eth_apis import (
    disabled_method_list,
)
from conflux_web3.types import (
    EstimateResult,
    FeeHistory,
    TxReceipt,
    TxData,
    NodeStatus,
    FilterParams,
    LogReceipt,
    BlockData,
    SponsorInfo,
    AccountInfo,
    DepositInfo,
    TxReceiptWithSpace,
    VoteInfo,
    StorageRoot,
    BlockRewardInfo,
    PendingInfo,
    PoSEconomicsInfo,
    PoSEpochRewardInfo,
    DAOVoteInfo,
    SupplyInfo,
    PendingTransactionsInfo,
    TransactionPaymentInfo,
    CollateralInfo,
    BlockFilterId,
    TxFilterId,
    LogFilterId,
    _FilterId,
)
from conflux_web3.contract import (
    ConfluxContract
)
from conflux_web3.contract.metadata import (
    get_contract_metadata
)
from conflux_web3._utils.transactions import (
    fill_transaction_defaults
)
from conflux_web3._utils.normalizers import (
    rpc_snake_to_camel
)
from conflux_web3.method import (
    ConfluxMethod
)
from conflux_web3.middleware.pending import (
    TransactionHash
)
from conflux_web3._utils.decorators import (
    use_instead,
)
from conflux_web3._utils.cns import (
    resolve_if_cns_name,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

class BaseCfx(BaseEth):
    _default_block: EpochNumberParam = "latest_state"
    _default_account: Union[AddressParam, Empty] = empty
    w3: "Web3"
    
    @property
    def default_block(self) -> EpochNumberParam:
        return self._default_block

    @default_block.setter
    def default_block(self, value: EpochNumberParam) -> None:
        self._default_block = value
    
    @property
    def default_account(self) -> Base32Address:
        """default account address, this address will be used as default `from` address if transaction `from` field is empty
        """
        return self._default_account # type: ignore
    

    @default_account.setter
    def default_account(self, account: Union[Base32Address, str, LocalAccount]) -> None:
        """set default account address
        """
        if isinstance(account, LocalAccount):
            normalized_address = Base32Address(account.address)
            self._default_account = normalized_address
            if (self.w3.wallet is not None and account.address not in self.w3.wallet): # type: ignore
                self.w3.wallet.add_account(account)
            if self.w3.cns:
                self.w3.cns.w3.cfx.default_account = normalized_address
        else:
            normalized_address = Base32Address(resolve_if_cns_name(self.w3, account))
            self._default_account = normalized_address
            if self.w3.cns:
                self.w3.cns.w3.cfx.default_account = normalized_address
    
    def remove_default_account(self) -> None:
        self._default_account = empty
        if self.w3.cns:
            self.w3.cns.w3.cfx.remove_default_account()
    
    def send_transaction_munger(self, transaction: TxParam) -> Tuple[TxParam]:
        if 'from' not in transaction and self.default_account:
            transaction = assoc(transaction, 'from', self.default_account)
        if 'value' in transaction:
            transaction['value'] = to_int_if_drip_units(transaction['value'])
        if 'gasPrice' in transaction:
            transaction['gasPrice'] = to_int_if_drip_units(transaction['gasPrice'])
        if 'maxPriorityFeePerGas' in transaction:
            transaction['maxPriorityFeePerGas'] = to_int_if_drip_units(transaction['maxPriorityFeePerGas'])
        if 'maxFeePerGas' in transaction:
            transaction['maxFeePerGas'] = to_int_if_drip_units(transaction['maxFeePerGas'])
        transaction = fill_transaction_defaults(self.w3, transaction)
        return (transaction,)
    
    def estimate_gas_and_collateral_munger(
        self, transaction: TxParam, block_identifier: Optional[EpochNumberParam]=None
    ) -> Tuple[TxParam, Optional[EpochNumberParam]]:
        if "from" not in transaction and self.default_account:
            transaction = assoc(transaction, "from", self.default_account)
        if 'value' in transaction:
            transaction['value'] = to_int_if_drip_units(transaction['value'])
        if 'gasPrice' in transaction:
            transaction['gasPrice'] = to_int_if_drip_units(transaction['gasPrice'])

        if block_identifier is None:
            params = (transaction, self._default_block)
        else:
            params = (transaction, block_identifier)

        return params
    
    def default_account_munger(
        self, address: Optional[Union[AddressParam, LocalAccount, Empty]]=None, block_identifier: Optional[EpochNumberParam]=None
    ) -> Sequence[Union[AddressParam, Empty, EpochNumberParam]]:
        if not address:
            if self.default_account:
                address = self.default_account
            else:
                raise ValueError(
                    "address parameter is required, set address or set 'web3.cfx.default_account'"
                )

        return [address, block_identifier] # type: ignore
    
    _clientVersion: ConfluxMethod[Callable[[], str]] = ConfluxMethod(
        RPC.cfx_clientVersion,
    )
    
    _get_status: ConfluxMethod[Callable[[], NodeStatus]] = ConfluxMethod(
        RPC.cfx_getStatus,
    )

    _gas_price: ConfluxMethod[Callable[[], Drip]] = ConfluxMethod(
        RPC.cfx_gasPrice,
    )
    
    
    _estimate_gas_and_collateral: ConfluxMethod[Callable[..., EstimateResult]] = ConfluxMethod(
        RPC.cfx_estimateGasAndCollateral, mungers=[estimate_gas_and_collateral_munger]
    )
    
    _accounts: ConfluxMethod[Callable[[], Tuple[Base32Address]]] = ConfluxMethod(
        RPC.accounts
    )
    
    _call: ConfluxMethod[Callable[[TxParam, Optional[EpochNumberParam]], Any]] = ConfluxMethod(
        RPC.cfx_call,
        mungers=[default_account_munger]
    )
    
    _get_balance: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Drip]] = ConfluxMethod(
        RPC.cfx_getBalance,
        # mungers=[default_account_munger]
    )
    
    _get_staking_balance: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Drip]] = ConfluxMethod(
        RPC.cfx_getStakingBalance
    )
    
    _epoch_number: ConfluxMethod[Callable[[Optional[EpochLiteral]], EpochNumber]] = ConfluxMethod(
        RPC.cfx_epochNumber,
    )
    
    _get_next_nonce: ConfluxMethod[Callable[..., int]] = ConfluxMethod(
        RPC.cfx_getNextNonce,
        # mungers=[default_account_munger]
    )
    
    _send_raw_transaction: ConfluxMethod[Callable[[Union[HexStr, bytes]], HexBytes]] = ConfluxMethod(
        RPC.cfx_sendRawTransaction
    )
    
    _send_transaction: ConfluxMethod[Callable[[TxParam], HexBytes]] = ConfluxMethod(
        RPC.cfx_sendTransaction,
        mungers=[send_transaction_munger],
    )
    
    _get_confirmation_risk_by_hash: ConfluxMethod[Callable[[_Hash32], float]] = ConfluxMethod(
        RPC.cfx_getConfirmationRiskByHash,
    )
    
    _get_transaction_receipt: ConfluxMethod[Callable[[_Hash32], TxReceipt]] = ConfluxMethod(
        RPC.cfx_getTransactionReceipt
    )
    
    _transaction_receipt: ConfluxMethod[Callable[[_Hash32], TxReceipt]] = ConfluxMethod(
        RPC.cfx_getTransactionReceipt
    )
    
    _get_transaction_by_hash: ConfluxMethod[Callable[[_Hash32], TxData]] = ConfluxMethod(
        RPC.cfx_getTransactionByHash
    )
    
    _get_block_by_hash: ConfluxMethod[Callable[[_Hash32, bool], BlockData]] = ConfluxMethod(
        RPC.cfx_getBlockByHash
    )
    
    _get_block_by_epoch_number: ConfluxMethod[Callable[[EpochNumberParam, bool], BlockData]] = ConfluxMethod(
        RPC.cfx_getBlockByEpochNumber
    )
    
    _get_block_by_block_number: ConfluxMethod[Callable[[int, bool], BlockData]] = ConfluxMethod(
        RPC.cfx_getBlockByBlockNumber
    )
    
    _get_best_block_hash: ConfluxMethod[Callable[[], HexBytes]] = ConfluxMethod(
        RPC.cfx_getBestBlockHash
    )
    
    _get_blocks_by_epoch: ConfluxMethod[Callable[[EpochNumberParam], Sequence[HexBytes]]] = ConfluxMethod(
        RPC.cfx_getBlocksByEpoch
    )
    
    _get_skipped_blocks_by_epoch: ConfluxMethod[Callable[[EpochNumberParam], Sequence[HexBytes]]] = ConfluxMethod(
        RPC.cfx_getSkippedBlocksByEpoch
    )
    
    _get_block_by_hash_with_pivot_assumptions: ConfluxMethod[Callable[[_Hash32, _Hash32, int], BlockData]] = ConfluxMethod(
        RPC.cfx_getBlockByHashWithPivotAssumption
    )

    _get_epoch_receipts: ConfluxMethod[Callable[[EpochNumberParam, Optional[bool]], Sequence[Sequence[TxReceipt]]]] = ConfluxMethod(RPC.cfx_getEpochReceipts)

    _get_code: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], HexBytes]] = ConfluxMethod(
        RPC.cfx_getCode
    )
    
    _get_storage_at: ConfluxMethod[Callable[[AddressParam, int, Optional[EpochNumberParam]], Union[HexBytes, None]]] = ConfluxMethod(
        RPC.cfx_getStorageAt
    )
    
    _get_storage_root: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Union[StorageRoot, None]]] = ConfluxMethod(
        RPC.cfx_getStorageRoot
    )
    
    _get_collateral_for_storage: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Storage]] = ConfluxMethod(
        RPC.cfx_getCollateralForStorage
    )
    
    _get_admin: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Union[Base32Address, None]]] = ConfluxMethod(
        RPC.cfx_getAdmin
    )
    
    _get_sponsor_info: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], SponsorInfo]] = ConfluxMethod(
        RPC.cfx_getSponsorInfo
    )
    
    _get_account: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], AccountInfo]] = ConfluxMethod(
        RPC.cfx_getAccount
    )
    
    _get_deposit_list: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Sequence[DepositInfo]]] = ConfluxMethod(
        RPC.cfx_getDepositList
    )
    
    _get_vote_list: ConfluxMethod[Callable[[AddressParam, Optional[EpochNumberParam]], Sequence[VoteInfo]]] = ConfluxMethod(
        RPC.cfx_getVoteList
    )
    
    _get_interest_rate = ConfluxMethod(
        RPC.cfx_getInterestRate
    )
    
    _get_accumulate_interest_rate = ConfluxMethod(
        RPC.cfx_getAccumulateInterestRate
    )
    
    _get_block_reward_info = ConfluxMethod(
        RPC.cfx_getBlockRewardInfo
    )
    
    _get_pos_economics = ConfluxMethod(
        RPC.cfx_getPoSEconomics
    )
    
    _get_pos_reward_by_epoch = ConfluxMethod(
        RPC.cfx_getPoSRewardByEpoch
    )
    
    _get_params_from_vote = ConfluxMethod(
        RPC.cfx_getParamsFromVote
    )
    
    _get_supply_info = ConfluxMethod(
        RPC.cfx_getSupplyInfo
    )
    
    _get_account_pending_info = ConfluxMethod(
        RPC.cfx_getAccountPendingInfo
    )
    
    _get_account_pending_transactions = ConfluxMethod(
        RPC.cfx_getAccountPendingTransactions
    )
    
    _check_balance_against_transaction = ConfluxMethod(
        RPC.cfx_checkBalanceAgainstTransaction
    )
    
    _max_priority_fee: ConfluxMethod[Callable[[], Drip]] = ConfluxMethod(
        RPC.cfx_maxPriorityFeePerGas
    )
    
    _fee_history: ConfluxMethod[Callable[[int, EpochNumberParam, Sequence[float]], FeeHistory]] = ConfluxMethod(
        RPC.cfx_feeHistory
    )
    
    _get_fee_burnt: ConfluxMethod[Callable[[], Drip]] = ConfluxMethod(
        RPC.cfx_getFeeBurnt
    )
    
    _get_logs: ConfluxMethod[Callable[[FilterParams], List[LogReceipt]]] = ConfluxMethod(
        RPC.cfx_getLogs
    )
    
    _new_filter: ConfluxMethod[Callable[[FilterParams], LogFilterId]] = ConfluxMethod(RPC.cfx_newFilter)

    _new_block_filter: ConfluxMethod[Callable[[], BlockFilterId]] = ConfluxMethod(RPC.cfx_newBlockFilter)
    
    _new_pending_transaction_filter: ConfluxMethod[Callable[[], TxFilterId]] = ConfluxMethod(RPC.cfx_newPendingTransactionFilter)

    _get_filter_changes: ConfluxMethod[Callable[[_FilterId], Union[Sequence[Hash32], Sequence[LogReceipt]]]] = ConfluxMethod(RPC.cfx_getFilterChanges)
    
    _get_filter_logs: ConfluxMethod[Callable[[Union[LogFilterId, str]], Sequence[LogReceipt]]] = ConfluxMethod(RPC.cfx_getFilterLogs)

    _uninstall_filter: ConfluxMethod[Callable[[_FilterId], bool]] = ConfluxMethod(RPC.cfx_uninstallFilter)

    _get_collateral_info: ConfluxMethod[Callable[[Optional[EpochNumberParam]], CollateralInfo]] = ConfluxMethod(RPC.cfx_getCollateralInfo)

    @overload  
    def contract(
        self, address: Union[Base32Address, str], *, name: Optional[str]=None, with_deployment_info: Optional[bool]=None, **kwargs: Any
    ) -> ConfluxContract:
        ...

    @overload
    def contract(
        self, address: None=None, *, name: None=None, with_deployment_info: Optional[bool]=None, **kwargs: Any
    ) -> Type[ConfluxContract]:
        ...
    
    @overload  
    def contract(
        self, address: None=None, *, name: str=..., with_deployment_info: None=None, **kwargs: Any  
    ) -> Union[Type[ConfluxContract], ConfluxContract]:
        ...
    
    @overload  
    def contract(
        self, address: None=None, *, name: str=..., with_deployment_info: Literal[False]=..., **kwargs: Any 
    ) -> Type[ConfluxContract]:
        ...

    @overload  
    def contract(
        self, address: None=None, *, name: str=..., with_deployment_info: Literal[True]=..., **kwargs: Any
    ) -> ConfluxContract:
        ...

    def contract(
        self,
        address: Optional[Union[Base32Address, str]] = None,
        *,
        name: Optional[str] = None,
        with_deployment_info: Optional[bool] = None,
        **kwargs: Any,
    ) -> Union[Type[ConfluxContract], ConfluxContract]:
        """
        Produce a contract factory (address is not specified) or a contract(address is not specified).
        Address is specified by:

            1. explicitly using address param
            2. embedded contract deployment info if 
                (1) "name" parameter specified and corresponding contract has deployment info 
                (2) with_deployment_info is not False

        >>> from conflux_web3.dev import get_mainnet_web3
        >>> w3 = get_mainnet_web3()
        >>> from conflux_web3.contract import get_contract_metadata
        >>> abi = get_contract_metadata("ERC20")["abi"]
        >>> bytecode = get_contract_metadata("ERC20")["bytecode"]
        >>> erc20_factory = w3.cfx.contract(abi=abi, bytecode=bytecode)

        >>> c1 = w3.cfx.contract(name="AdminControl")
        >>> assert c1.address # address is added by default
        >>> c2 = w3.cfx.contract(name="AdminControl", with_deployment_info=False)
        >>> assert not c2.address # address is not added by explicitly specified

        Parameters
        ----------
        address : Optional[Union[Base32Address, str]], optional
            the address of the contract, by default None
        name : Optional[str], optional
            the name of the contract, which is used to specify abi, bytecode, or deployed contract address, by default None
        with_deployment_info : Optional[bool], optional
            whether address will be specified if name parameter is provided, 
            if True, the address will always be specified if name argument is provided
            if False, the address will never be specified if name argument is provided
            if None, the address will be specified depending on if corresponding address exists
            by default None
        **kwargs: Dict[str, Any]
            used to specify abi and bytecode argument
        Returns
        -------
        Union[Type[ConfluxContract], ConfluxContract]
            returns a contract factory or contract
        """        
        metadata = {}
        if name is not None:
            metadata = get_contract_metadata(name, self.chain_id, with_deployment_info) # type: ignore
            # the latter one shares greater priority when merging
            kwargs = merge(metadata, kwargs)
        if address is not None:
            kwargs["address"] = address
        return super().contract(**kwargs)  # type: ignore

class ConfluxClient(BaseCfx, Eth):
    """RPC entry defined provides friendlier APIs for users
    """
    account: Account
    _default_contract_factory: Type[ConfluxContract] = ConfluxContract
    _allow_arbitary_rpc: bool = False

    def __init__(self, w3: "Web3") -> None:
        super().__init__(w3)
        self.account = Account()
        self.account.set_w3(w3)
        self._disable_eth_methods(disabled_method_list)

    def __getattr__(self, name: str) -> Callable[..., Any]:
        if not self._allow_arbitary_rpc:
            raise Exception(f"RPC method cfx_{rpc_snake_to_camel(name)} is not defined, you can set `web3.cfx._allow_arbitary_rpc=True` to enable any RPC call.")
        method: ConfluxMethod[Callable[..., Any]] = ConfluxMethod(RPCEndpoint(f"cfx_{rpc_snake_to_camel(name)}"))
        return method.__get__(self, self.__class__)

    # lazy initialize self.address
    @cached_property
    def address(self) -> Type[Base32Address]:
        return get_base32_address_factory(self.chain_id)
        
    def _disable_eth_methods(self, disabled_method_list: Sequence[str]):
        for api in disabled_method_list:
            always_returns_zero: Callable[..., Literal[0]] = lambda *args, **kwargs: 0
            self.__setattr__(
                api,
                use_instead(origin=api)(
                    always_returns_zero
                ),
            )
            
    @property
    @use_instead
    def syncing(self):
        """
        Unsupported API
        """
        pass
    
    @property
    @use_instead
    def coinbase(self):
        """
        # WARNING: Unsupported API
        """
        pass

    
    @property
    @use_instead
    def mining(self):
        """
        # WARNING: Unsupported API
        """
        pass
    
    @property
    @use_instead
    def hashrate(self):
        """
        # WARNING: Unsupported API
        """
        pass
    
    @property
    @use_instead(origin="web3.eth.block_number", substitute="web3.cfx.epoch_number")
    def block_number(self):
        """
        # WARNING: Unsupported API, use `web3.cfx.epoch_number` instead
        """
        pass
    
    @property
    @use_instead
    def get_work(self):
        """
        # WARNING: Unsupported API
        """
        pass
    
    def get_status(self) -> NodeStatus:
        """
        get the blockchain status from the provider
        
        >>> w3.cfx.get_status()
        AttributeDict({'bestHash': HexBytes('0xb91109cc301209921b33533f8ae228615016b9104e8229ddaad55b370fe586d3'),
            'chainId': 1,
            'ethereumSpaceChainId': 71,
            'networkId': 1,
            'epochNumber': 97105091,
            'blockNumber': 124293823,
            'pendingTxNumber': 69,
            'latestCheckpoint': 97000000,
            'latestConfirmed': 97105034,
            'latestState': 97105087,
            'latestFinalized': 97104660})
        >>> w3.cfx.get_status().chainId # support but not recommended because of missing type hints
        1
        >>> w3.cfx.get_status()['chainId'] # recommended
        1

        Returns
        -------
        NodeStatus
            a dict representing node status
        """        
        return self._get_status()
    
    @property
    def gas_price(self) -> Drip:
        """
        Get the current gas price
        
        >>> w3.cfx.gas_price
        1000000000 Drip
        >>> w3.cfx.gas_price.value
        1000000000
        >>> w3.cfx.gas_price.to("GDrip")
        1 GDrip

        Returns
        -------
        Drip
            gas price wrapped by Drip, which can be used as transaction value or gas price
        """        
        return self._gas_price()
    
    @property
    def accounts(self) -> Tuple[Base32Address]:
        """
        Returns the accounts controlled by the provider 
        ## this method is not supported by public rpcs

        Returns
        -------
        Tuple[Base32Address]
            a Tuple containing the address of the accounts
        """
        return self._accounts()
    
    @property
    def epoch_number(self) -> EpochNumber:
        """
        Returns latest_mined epoch number

        Returns
        -------
        EpochNumber
            latest mined epoch number
        """
        return self._epoch_number(None)
    
    def epoch_number_by_tag(self, epoch_tag: EpochLiteral) -> EpochNumber:
        """
        Returns the epoch numebr with the provided epoch tag

        Parameters
        ----------
        epoch_tag : EpochLiteral
            String "latest_mined", "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest"

        Returns
        -------
        EpochNumber
            the epoch number corrensponding to the epoch tag
        """        
        return self._epoch_number(epoch_tag)
    
    @cached_property
    def chain_id(self) -> int:
        """
        Get the chain id of the current network.
        This property is cached and won't be changed unless it is deleted

        Returns
        -------
        int
            chain id of the blockchain, 1 for conflux testnet and 1029 for conflux mainnet
        """
        return self._get_status()["chainId"]

    @property
    def client_version(self):
        return self._clientVersion()
    
    # def is_connected(self):
    #     return 
    
    def get_balance(self,
                    address: Union[Base32Address, str], 
                    block_identifier: Optional[EpochNumberParam] = None) -> Drip:
        """
        Returns the balance of the given account, identified by its address.

        >>> balance = w3.cfx.get_balance("cfx:type.user:aarc9abycue0hhzgyrr53m6cxedgccrmmyybjgh4xg")
        >>> balance
        "1000000000000000000 Drip"
        >>> balance.value
        1000000000000000000
        >>> balance.to("CFX")
        "1 CFX"

        Parameters
        ----------
        address : Union[Base32Address, str]
            the address of the account 
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        Drip
            balance wrapped by Drip
        """        
        return self._get_balance(address, block_identifier)
    
    def get_staking_balance(self,
                    address: Union[Base32Address, str], 
                    block_identifier: Optional[EpochNumberParam] = None) -> Drip:
        """
        Returns the staking balance of the given account, identified by its address.

        Parameters
        ----------
        address : Union[Base32Address, str]
            the address of the account 
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        Drip
            balance wrapped by Drip
        """        
        return self._get_staking_balance(address, block_identifier)
    
    
    def call(self, 
             transaction: TxParam, 
             block_identifier: Optional[EpochNumberParam]=None, 
             **kwargs: Dict[str, Any]) -> Any:
        """
        Virtually calls a contract, returns the output data. The transaction will not be added to the blockchain.

        Parameters
        ----------
        transaction : TxParam
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        Any
        """        
        return self._call(transaction, block_identifier)
    
    def get_next_nonce(self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None) -> int:
        """
        Returns the next nonce that should be used by the given account when sending a transaction.

        Parameters
        ----------
        address : Union[Base32Address, str]
            the address of the account
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        int
        """        
        return self._get_next_nonce(address, block_identifier)

    def get_transaction_count(self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None) -> int:
        """
        Compatibility API for conflux. Equivalent to `get_next_nonce`
        """        
        return self.get_next_nonce(address, block_identifier)

    def estimate_gas_and_collateral(self, transaction: TxParam, block_identifier: Optional[EpochNumberParam]=None) -> EstimateResult:
        """
        Virtually executes a transaction, returns an estimate for the size of storage collateralized and the gas used by the transaction. The transaction will not be added to the blockchain.
        For most cases, you don't need to invoke this api to fill the transaction fields unless you need to manually sign the transactions.

        >>> w3.cfx.estimate_gas_and_collateral({"to": "cfx:type.contract:acc7uawf5ubtnmezvhu9dhc6sghea0403y2dgpyfjp"})
        >>> {"gasLimit": 21000, "gasUsed": 21000, "storageCollateralized": 0}

        Parameters
        ----------
        transaction: TxParam
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        EstimateResult
        """        
        return self._estimate_gas_and_collateral(transaction, block_identifier)

    def estimate_gas(self, transaction: TxParam, block_identifier: Optional[EpochNumberParam] = None) -> EstimateResult:
        """
        Compatibility API for conflux. Equivalent to `estimate_gas_and_collateral`
        """        
        return self.estimate_gas_and_collateral(transaction, block_identifier)

    def send_raw_transaction(self, raw_transaction: Union[HexStr, bytes]) -> TransactionHash:
        """
        Sends a signed transaction into the network for processing
        
        >>> tx_hash = w3.cfx.send_raw_transaction("0xf86eea8201a28207d0830f4240943838197c0c88d0d5b13b67e1bfdbdc132d4842e389056bc75e2d631000008080a017b8b26f473820475edc49bd153660e56b973b5985bbdb2828fceacb4c91f389a03452f9a69da34ef35acc9c554d7b1d63e9041141674b42c3abb1b57b9f83a2d3")
        >>> tx_hash.executed() # equivalent to w3.cfx.wait_for_transaction_receipt(tx_hash)

        Parameters
        ----------
        raw_transaction : Union[HexStr, bytes]
            a signed transaction as hex string or bytes

        Returns
        -------
        TransactionHash
            the hash of the transaction in bytes and wrapped by TransactionHash object which provides chained operations
        """        
        # TODO: remove pending middleware and changes here
        return cast(TransactionHash, self._send_raw_transaction(raw_transaction))
    
    def send_transaction(self, transaction: TxParam) -> TransactionHash:
        """
        Send a transaction by using transaction params. 
        ## The node MUST support cfx_sendTransaction rpc (which is typically not supported by public rpc services)
        ## or account is added to w3.wallet
        
        >>> account = w3.account.create()
        >>> w3.wallet.add_account(account)
        >>> w3.cfx.send_transaction({
        ...    "to": w3.address.zero_address(),
        ...    "from": account.address,
        ...    "value": 100,
        ... }).executed()

        Parameters
        ----------
        transaction : TxParam

        Returns
        -------
        TransactionHash
            the hash of the transaction in bytes and wrapped by TransactionHash object which provides chained operations
        """        
        # TODO: remove pending middleware and changes here
        return cast(TransactionHash, self._send_transaction(transaction))
    
    def get_transaction_receipt(self, transaction_hash: _Hash32) -> TxReceipt:
        """
        Returns a transaction receipt, identified by the corresponding transaction hash.
        
        >>> w3.cfx.get_transaction_receipt("0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980")
        AttributeDict({'transactionHash': HexBytes('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980'),
            'index': 227,
            'blockHash': HexBytes('0xceda961d541c78fa2c99907620bb2b13707f72b004b7c70362a5642177e6aae2'),
            'epochNumber': 97108737,
            'from': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'to': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'gasUsed': 21000,
            'gasFee': 21000000000000 Drip,
            'contractCreated': None,
            'logs': [],
            'logsBloom': HexBytes('0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'),
            'stateRoot': HexBytes('0x89dd70006e4714eb81a210a20badf4c37adc7fcbe992b8458460876fcac38ea3'),
            'outcomeStatus': 0,
            'txExecErrorMsg': None,
            'gasCoveredBySponsor': False,
            'storageCoveredBySponsor': False,
            'storageCollateralized': 0,
            'storageReleased': []})

        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string

        Returns
        -------
        TxReceipt
            the receipt of the transaction
        """        
        return self._get_transaction_receipt(transaction_hash)
    
    def wait_till_transaction_mined(
        self, transaction_hash: _Hash32, timeout: float = 60, poll_latency: float = 0.5
    ) -> TxData:
        """
        Returns information about a transaction when it is mined.
        If TxData is returned, the transaction is found contained in a block, but might not be executed.
        (Transaction will only be executed after 5 epochs)
        
        >>> w3.cfx.wait_till_transaction_mined("0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980")
        AttributeDict({'hash': HexBytes('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980'),
            'nonce': 8230,
            'blockHash': HexBytes('0xceda961d541c78fa2c99907620bb2b13707f72b004b7c70362a5642177e6aae2'),
            'transactionIndex': 227,
            'from': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'to': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'value': 0 Drip,
            'gasPrice': 1000000000 Drip,
            'gas': 21000,
            'contractCreated': None,
            'data': HexBytes('0x'),
            'storageLimit': 0,
            'epochHeight': 97108719,
            'chainId': 1,
            'status': 0,
            'v': 1,
            'r': HexBytes('0x1d16cb28acf3973df4f4cc29ffffdba5d36ca1e38f1904f64780fa505691c1b7'),
            's': HexBytes('0x1549267bc4f60af38d23112b47a54dc36481e77eb44dd5ee2c3277810c1c6297')})


        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string
        timeout : float, optional
            maximum wait time before timeout in seconds, by default 60
        poll_latency : float, optional
            time interval to query transaction status in seconds, by default 0.5

        Returns
        -------
        TxData
            the transaction data as a dict

        Raises
        ------
        TimeExhausted
            if timeout
        """        
        try:
            with Timeout(timeout) as _timeout:
                while True:
                    try:
                        tx_data = self.get_transaction_by_hash(transaction_hash)
                    except TransactionNotFound:
                        tx_data = None
                    if tx_data is not None and tx_data["blockHash"]:
                        break
                    _timeout.sleep(poll_latency)
            return tx_data

        except Timeout:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not in the chain "
                f"after {timeout} seconds"
            )
    
    def wait_for_transaction_receipt(
        self, transaction_hash: _Hash32, timeout: float = 300, poll_latency: float = 0.5
    ) -> TxReceipt:
        """
        Alias for `wait_till_transaction_executed`
        """
        try:
            receipt = cast(TxReceipt, super().wait_for_transaction_receipt(transaction_hash, timeout, poll_latency)) # type: ignore
        except TimeExhausted:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not executed"
                f"after {timeout} seconds"
            )
        if receipt["outcomeStatus"] != 0:
            raise RuntimeError(f'transaction "${transaction_hash}" execution failed, outcomeStatus ${receipt["outcomeStatus"]}')
        return receipt
    
    def wait_till_transaction_executed(
        self, transaction_hash: _Hash32, timeout: float = 300, poll_latency: float = 0.5
    ) -> TxReceipt:
        """
        Returns transaction receipt after it is executed.
        If TxReceipt is returned, the transaction is found executed successfully
        
        >>> w3.cfx.wait_till_transaction_executed("0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980")
        AttributeDict({'transactionHash': HexBytes('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980'),
            'index': 227,
            'blockHash': HexBytes('0xceda961d541c78fa2c99907620bb2b13707f72b004b7c70362a5642177e6aae2'),
            'epochNumber': 97108737,
            'from': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'to': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'gasUsed': 21000,
            'gasFee': 21000000000000 Drip,
            'contractCreated': None,
            'logs': [],
            'logsBloom': HexBytes('0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'),
            'stateRoot': HexBytes('0x89dd70006e4714eb81a210a20badf4c37adc7fcbe992b8458460876fcac38ea3'),
            'outcomeStatus': 0,
            'txExecErrorMsg': None,
            'gasCoveredBySponsor': False,
            'storageCoveredBySponsor': False,
            'storageCollateralized': 0,
            'storageReleased': []})

        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string
        timeout : float, optional
            maximum wait time before timeout in seconds, by default 300
        poll_latency : float, optional
            time interval to query transaction status in seconds, by default 0.5

        Returns
        -------
        TxReceipt
            transaction receipt as a dict

        Raises
        ------
        TimeExhausted
            if timeout
        RuntimeError
            if transaction is not executed successfully
        """  
        return self.wait_for_transaction_receipt(transaction_hash, timeout, poll_latency)
        
    
    def wait_till_transaction_confirmed(
        self, transaction_hash: _Hash32, timeout: float = 600, poll_latency: float = 0.5
    ) -> TxReceipt:
        """
        Returns transaction receipt after a transaction is confirmed.
        If TxReceipt is returned, the chances that the transaction execution result to be reverted is tiny.
        But there are still chances that the transaction might be reverted.
        ## Developers should use `wait_till_transaction_finalized` if the transaction is REALLY IMPORTANT.
        
        >>> w3.cfx.get_transaction_receipt("0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980")
        AttributeDict({'transactionHash': HexBytes('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980'),
            'index': 227,
            'blockHash': HexBytes('0xceda961d541c78fa2c99907620bb2b13707f72b004b7c70362a5642177e6aae2'),
            'epochNumber': 97108737,
            'from': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'to': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'gasUsed': 21000,
            'gasFee': 21000000000000 Drip,
            'contractCreated': None,
            'logs': [],
            'logsBloom': HexBytes('0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'),
            'stateRoot': HexBytes('0x89dd70006e4714eb81a210a20badf4c37adc7fcbe992b8458460876fcac38ea3'),
            'outcomeStatus': 0,
            'txExecErrorMsg': None,
            'gasCoveredBySponsor': False,
            'storageCoveredBySponsor': False,
            'storageCollateralized': 0,
            'storageReleased': []})

        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string
        timeout : float, optional
            maximum wait time before timeout in seconds, by default 600
        poll_latency : float, optional
            time interval to query transaction status in seconds, by default 0.5

        Returns
        -------
        TxReceipt
            transaction receipt as a dict

        Raises
        ------
        TimeExhausted
            if timeout
        RuntimeError
            if transaction is not executed successfully
        """  
        try:
            with Timeout(timeout) as _timeout:
                while True:
                    try:
                        tx_receipt = self.wait_till_transaction_executed(transaction_hash)
                    except TransactionNotFound:
                        tx_receipt = None
                    if tx_receipt is not None:
                        tx_epoch = tx_receipt["epochNumber"]
                        confirmed_epoch = self.epoch_number_by_tag("latest_confirmed")
                        if tx_epoch <= confirmed_epoch:
                            break
                    _timeout.sleep(poll_latency)
            return tx_receipt
        except Timeout:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not confirmed "
                f"after {timeout} seconds"
            )
    
    def wait_till_transaction_finalized(
        self, transaction_hash: _Hash32, timeout: float = 1200, poll_latency: float = 5
    ) -> TxReceipt:
        """
        Returns transaction receipt after a transaction is finalized by PoS chain.
        It will take 4 ～ 6 minutes to finalize a transaction but the finalized transaction won't be reverted.
        It is recommended for developers to use this api to confirm important transactions.

        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string
        timeout : float, optional
            maximum wait time before timeout in seconds, by default 1200
        poll_latency : float, optional
            time interval to query transaction status in seconds, by default 0.5

        Returns
        -------
        TxReceipt
            transaction receipt as a dict

        Raises
        ------
        TimeExhausted
            If timeout
        """        
        warnings.warn("4 ~ 6 minutes are required to finalize a transaction", UserWarning)
        try:
            tx_receipt = self.wait_till_transaction_executed(transaction_hash)
            with Timeout(timeout) as _timeout:
                while True:
                    tx_epoch = tx_receipt["epochNumber"]
                    finalized_epoch = self.epoch_number_by_tag("latest_finalized")
                    if tx_epoch <= finalized_epoch:
                        break
                    _timeout.sleep(poll_latency)
            return tx_receipt
        except Timeout:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not finalized "
                f"after {timeout} seconds"
            )
    
    def get_transaction_by_hash(self, transaction_hash: _Hash32) -> TxData:
        """
        Returns information about a transaction, identified by its hash.
        ## Note: the transaction is not mined if `blockHash` field is None
        
        >>> w3.cfx.get_transaction_by_hash('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980')
        AttributeDict({'hash': HexBytes('0x5ffa11a44c6db42cc30967d4de5949b17ee319c4aba72f3380c60136993a8980'),
            'nonce': 8230,
            'blockHash': HexBytes('0xceda961d541c78fa2c99907620bb2b13707f72b004b7c70362a5642177e6aae2'),
            'transactionIndex': 227,
            'from': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'to': 'cfxtest:aaksxj04phv0hp02jh3cbym9vghwkzap867jagb043',
            'value': 0 Drip,
            'gasPrice': 1000000000 Drip,
            'gas': 21000,
            'contractCreated': None,
            'data': HexBytes('0x'),
            'storageLimit': 0,
            'epochHeight': 97108719,
            'chainId': 1,
            'status': 0,
            'v': 1,
            'r': HexBytes('0x1d16cb28acf3973df4f4cc29ffffdba5d36ca1e38f1904f64780fa505691c1b7'),
            's': HexBytes('0x1549267bc4f60af38d23112b47a54dc36481e77eb44dd5ee2c3277810c1c6297')})


        Parameters
        ----------
        transaction_hash : _Hash32
            the hash of the transaction, could be byte or hex string

        Returns
        -------
        TxData
            transaction data as a dict
        """        
        return self._get_transaction_by_hash(transaction_hash)
    
    # same as web3.py
    def get_transaction(self, transaction_hash: _Hash32) -> TxData:
        """Alias for `get_transaction_by_hash`
        """
        return self.get_transaction_by_hash(transaction_hash)
    
    def get_block_by_hash(
        self, block_hash: _Hash32, full_transactions: bool=False
    ) -> Union[BlockData, None]:
        """
        Returns information about a block, identified by its hash.

        >>> w3.cfx.get_block_by_hash("0x4752f93f61b43e3b09bcbcd36871d4894ac64254df367fa9d7a6253cfb618ece")
        AttributeDict({'hash': HexBytes('0x4752f93f61b43e3b09bcbcd36871d4894ac64254df367fa9d7a6253cfb618ece'),
            'parentHash': HexBytes('0x06a695ab273255341f24f955ce696d45007df4df66f0c61b1818965926c2e808'),
            'height': 97111729,
            'miner': 'cfxtest:aanpu16mtgc7dke5xhuktyfyef8f00pz8a2z5mc14g',
            'deferredStateRoot': HexBytes('0xeaf792fff684ce3f04c84614442ab38e8bb4622bacba1c01e857449ef97425f2'),
            'deferredReceiptsRoot': HexBytes('0x09f8709ea9f344a810811a373b30861568f5686e649d6177fd92ea2db7477508'),
            'deferredLogsBloomHash': HexBytes('0xd397b3b043d87fcd6fad1291ff0bfd16401c274896d8c63a923727f077b8e0b5'),
            'blame': 0,
            'transactionsRoot': HexBytes('0x6058a7cf739d94882638cada6e15a4f3b53b757c7195c5cafd04c2061f6de00d'),
            'epochNumber': 97111729,
            'blockNumber': 124302546,
            'gasLimit': 30000000,
            'gasUsed': 75291,
            'timestamp': 1666755747,
            'difficulty': 57822922,
            'powQuality': HexBytes('0x04ebbe82'),
            'refereeHashes': [],
            'adaptive': False,
            'nonce': HexBytes('0x14b30ccfa7cf95b2'),
            'size': 311,
            'custom': [HexBytes('0x02')],
            'posReference': HexBytes('0x2bac96d8e4205badeab4138d1f7b671a436a68ab50c54e7b154243cbf0c9205b'),
            'transactions': [HexBytes('0x70d17ef32c0544075078cb1ca777dab3728d443f00826cf3d05b9c01b76b5f5c')]})
                    
        Parameters
        ----------
        block_hash : _Hash32
            hash of a block
        full_transactions : bool, optional
            if true, it returns the full transaction objects. If false, only the hashes of the transactions are returned, by default False

        Returns
        -------
        Union[BlockData, None]
            the block data as a dict, or None if no block was found
        """        
        return self._get_block_by_hash(block_hash, full_transactions)
    
    def get_block_by_epoch_number(
        self, epoch_number_param: EpochNumberParam, full_transactions: bool=False
    ) -> BlockData:
        """
        Returns information about a pivot block, identified by its epoch number.

        >>> w3.cfx.get_block_by_epoch_number(97111729)
        AttributeDict({'hash': HexBytes('0x4752f93f61b43e3b09bcbcd36871d4894ac64254df367fa9d7a6253cfb618ece'),
            'parentHash': HexBytes('0x06a695ab273255341f24f955ce696d45007df4df66f0c61b1818965926c2e808'),
            'height': 97111729,
            'miner': 'cfxtest:aanpu16mtgc7dke5xhuktyfyef8f00pz8a2z5mc14g',
            'deferredStateRoot': HexBytes('0xeaf792fff684ce3f04c84614442ab38e8bb4622bacba1c01e857449ef97425f2'),
            'deferredReceiptsRoot': HexBytes('0x09f8709ea9f344a810811a373b30861568f5686e649d6177fd92ea2db7477508'),
            'deferredLogsBloomHash': HexBytes('0xd397b3b043d87fcd6fad1291ff0bfd16401c274896d8c63a923727f077b8e0b5'),
            'blame': 0,
            'transactionsRoot': HexBytes('0x6058a7cf739d94882638cada6e15a4f3b53b757c7195c5cafd04c2061f6de00d'),
            'epochNumber': 97111729,
            'blockNumber': 124302546,
            'gasLimit': 30000000,
            'gasUsed': 75291,
            'timestamp': 1666755747,
            'difficulty': 57822922,
            'powQuality': HexBytes('0x04ebbe82'),
            'refereeHashes': [],
            'adaptive': False,
            'nonce': HexBytes('0x14b30ccfa7cf95b2'),
            'size': 311,
            'custom': [HexBytes('0x02')],
            'posReference': HexBytes('0x2bac96d8e4205badeab4138d1f7b671a436a68ab50c54e7b154243cbf0c9205b'),
            'transactions': [HexBytes('0x70d17ef32c0544075078cb1ca777dab3728d443f00826cf3d05b9c01b76b5f5c')]})

        Parameters
        ----------
        epoch_number_param : EpochNumberParam
            the epoch number, or the string "latest_mined", "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest"
        full_transactions : bool, optional
            if true, it returns the full transaction objects. If false, only the hashes of the transactions are returned, by default False

        Returns
        -------
        BlockData
            the block data as a dict
        """        
        return self._get_block_by_epoch_number(epoch_number_param, full_transactions)
    
    def get_block_by_block_number(
        self, block_number: int, full_transactions: bool=False
    ) -> BlockData:
        """
        Returns information about a block, identified by its block number.

        >>> w3.cfx.get_block_by_block_number(124302546)
        AttributeDict({'hash': HexBytes('0x4752f93f61b43e3b09bcbcd36871d4894ac64254df367fa9d7a6253cfb618ece'),
            'parentHash': HexBytes('0x06a695ab273255341f24f955ce696d45007df4df66f0c61b1818965926c2e808'),
            'height': 97111729,
            'miner': 'cfxtest:aanpu16mtgc7dke5xhuktyfyef8f00pz8a2z5mc14g',
            'deferredStateRoot': HexBytes('0xeaf792fff684ce3f04c84614442ab38e8bb4622bacba1c01e857449ef97425f2'),
            'deferredReceiptsRoot': HexBytes('0x09f8709ea9f344a810811a373b30861568f5686e649d6177fd92ea2db7477508'),
            'deferredLogsBloomHash': HexBytes('0xd397b3b043d87fcd6fad1291ff0bfd16401c274896d8c63a923727f077b8e0b5'),
            'blame': 0,
            'transactionsRoot': HexBytes('0x6058a7cf739d94882638cada6e15a4f3b53b757c7195c5cafd04c2061f6de00d'),
            'epochNumber': 97111729,
            'blockNumber': 124302546,
            'gasLimit': 30000000,
            'gasUsed': 75291,
            'timestamp': 1666755747,
            'difficulty': 57822922,
            'powQuality': HexBytes('0x04ebbe82'),
            'refereeHashes': [],
            'adaptive': False,
            'nonce': HexBytes('0x14b30ccfa7cf95b2'),
            'size': 311,
            'custom': [HexBytes('0x02')],
            'posReference': HexBytes('0x2bac96d8e4205badeab4138d1f7b671a436a68ab50c54e7b154243cbf0c9205b'),
            'transactions': [HexBytes('0x70d17ef32c0544075078cb1ca777dab3728d443f00826cf3d05b9c01b76b5f5c')]})

        Parameters
        ----------
        epoch_number_param : EpochNumberParam
            the epoch number, or the string "latest_mined", "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest"
        full_transactions : bool, optional
            if true, it returns the full transaction objects. If false, only the hashes of the transactions are returned, by default False

        Returns
        -------
        BlockData
            the block data as a dict
        """  
        return self._get_block_by_block_number(block_number, full_transactions)
    
    def get_best_block_hash(self) -> HexBytes:
        """
        Returns the best block hash, which is the block hash of the latest pivot block

        Returns
        -------
        HexBytes
            best block hash in bytes
        """        
        return self._get_best_block_hash()
    
    def get_blocks_by_epoch(self, epoch_number_param: EpochNumberParam) -> Sequence[HexBytes]:
        """
        Returns the block hashes in the specified epoch.

        Parameters
        ----------
        epoch_number_param : EpochNumberParam
            the epoch number, or the string "latest_mined", "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest"

        Returns
        -------
        Sequence[HexBytes]
            a list of block hash in bytes
        """        
        return self._get_blocks_by_epoch(epoch_number_param)
    
    def get_skipped_blocks_by_epoch(self, epoch_number_param: EpochNumberParam) -> Sequence[HexBytes]:
        """
        Returns the list of non-executed blocks in an epoch. By default, Conflux only executes the last 200 blocks in each epoch (note that under normal circumstances, epochs should be much smaller).

        Parameters
        ----------
        epoch_number_param : EpochNumberParam
            the epoch number, or the string "latest_mined", "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest"

        Returns
        -------
        Sequence[HexBytes]
            a list of block hash in bytes
        """        
        return self._get_skipped_blocks_by_epoch(epoch_number_param)

    def get_block_by_hash_with_pivot_assumptions(self, block_hash: _Hash32, assumed_pivot_hash: _Hash32, epoch_number: int) -> BlockData:
        """
        Returns the requested block if the provided pivot hash is correct, returns an error otherwise.

        Parameters
        ----------
        block_hash : _Hash32
            block hash
        assumed_pivot_hash : _Hash32
            assumed pivot hash
        epoch_number : int
            integer epoch number

        Returns
        -------
        BlockData
            returns block data if pivot hash is correct
        """        
        return self._get_block_by_hash_with_pivot_assumptions(block_hash, assumed_pivot_hash, epoch_number)
    
    @overload
    def get_epoch_receipts(self, epoch_number: EpochNumberParam, include_espace_receipts: Literal[True]) -> Sequence[Sequence[TxReceiptWithSpace]]: ...

    @overload
    def get_epoch_receipts(self, epoch_number: EpochNumberParam, include_espace_receipts: Optional[Literal[False]]) -> Sequence[Sequence[TxReceipt]]: ...

    def get_epoch_receipts(self, epoch_number: EpochNumberParam, include_espace_receipts: Optional[bool]=False) -> Sequence[Sequence[Union[TxReceipt, TxReceiptWithSpace]]]:
        """
        Returns the transaction receipts in every block of the specific epoch. Returns 

        Parameters
        ----------
        epoch_number : EpochNumberParam
            the epoch to query, should be earlier than latest_state
        include_espace_receipts : Optional[bool]
            Defaults to False. Determines if the returned receipts will include eSpace receipts. 

        Returns
        -------
        Sequence[Sequence[TxReceipt]]
            returns a 2-dimension array containing block receipts, where transaction receipts are contained
        """
        return self._get_epoch_receipts(epoch_number, include_espace_receipts)
    
    def get_confirmation_risk_by_hash(self, block_hash: _Hash32) -> float:
        """
        Returns the confirmation risk of a given block, identified by its hash

        Parameters
        ----------
        block_hash : _Hash32
            block hash

        Returns
        -------
        float
            the confirmation risk in float
        """        
        return self._get_confirmation_risk_by_hash(block_hash)
    
    def get_block(self, block_identifier: Union[_Hash32, EpochNumberParam], full_transactions: bool = False) -> BlockData:
        """
        a simple wrapper combining `get_block_by_hash` and `get_block_by_epoch_number`
        # note: DON'T USE *BLOCK NUMBER* as the parameter. A number will be considered as epoch number
        
        >>> block_data = w3.cfx.get_block(97112003)
        >>> block_data_ = w3.cfx.get_block("0x1aa5c0d8d6778bd72ac842a52fd26042458b4f70a4d969dfa0d80f81f5eac790")
        >>> assert block_data == block_data_

        Parameters
        ----------
        block_identifier : Union[_Hash32, EpochNumberParam]
            block hash or epoch number parameter(e.g. string "latest_state" or epoch number)
        full_transactions : bool, optional
            whether the return block data contains full transaction info or just transaction hash, by default False

        Returns
        -------
        BlockData
        """
        # TODO: more fine-grained munger
        if block_identifier == "latest":
            block_identifier = "latest_state"  
        elif isinstance(block_identifier, bytes) or is_hash32_str(block_identifier):
            return self.get_block_by_hash(block_identifier, full_transactions) # type: ignore
        return self.get_block_by_epoch_number(block_identifier, full_transactions) # type: ignore
    
    def get_code(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> HexBytes:
        """
        Returns the code of the specified contract. If contract does not exist will return :const:`0x0`

        Parameters
        ----------
        address : Union[Base32Address, str]
            address of the contract
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", 
            "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        HexBytes
            byte code of the contract, or :const:`0x` if the account has no code
        """        
        return self._get_code(address, block_identifier)
    
    def get_storage_at(
        self, address: Union[Base32Address, str], storage_position: int, block_identifier: Optional[EpochNumberParam] = None
    ) -> Union[HexBytes, None]:
        """
        Returns storage entries from a given contract.

        Parameters
        ----------
        address : Union[Base32Address, str]
            address of the contract
        storage_position : int
            a storage position (refer to https://solidity.readthedocs.io/en/v0.7.1/internals/layout_in_storage.html for more info)
        block_identifier : Optional[EpochNumberParam], optional
            integer epoch number, or the string "latest_state", "latest_confirmed", "latest_checkpoint" or "earliest", by default None

        Returns
        -------
        Union[HexBytes, None]
            32 Bytes - the contents of the storage position, or :const:`None` if the contract does not exist
        """        
        return self._get_storage_at(address, storage_position, block_identifier)
    
    def get_storage_root(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> Union[StorageRoot, None]:
        return self._get_storage_root(address, block_identifier)
    
    def get_collateral_for_storage(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> Storage:
        return self._get_collateral_for_storage(address, block_identifier)
    
    def get_sponsor_info(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> SponsorInfo:
        return self._get_sponsor_info(address, block_identifier)
    
    def get_account(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> AccountInfo:
        return self._get_account(address, block_identifier)
    
    def get_deposit_list(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> Sequence[DepositInfo]:
        return self._get_deposit_list(address, block_identifier)
    
    def get_vote_list(
        self, address: Union[Base32Address, str], block_identifier: Optional[EpochNumberParam] = None
    ) -> Sequence[VoteInfo]:
        return self._get_vote_list(address, block_identifier)
    
    def get_interest_rate(
        self, block_identifier: Optional[EpochNumberParam] = None
    ) -> int:
        return self._get_interest_rate(block_identifier)
    
    def get_accumulate_interest_rate(
        self, block_identifier: Optional[EpochNumberParam] = None
    ) -> int:
        return self._get_accumulate_interest_rate(block_identifier)
    
    def get_block_reward_info(
        self, block_identifier: Union[EpochNumber, int, Literal["latest_checkpoint"] ]
    ) -> Sequence[BlockRewardInfo]:
        return self._get_block_reward_info(block_identifier)
    
    def get_pos_economics(
        self, block_identifier: Optional[EpochNumberParam] = None
    ) -> PoSEconomicsInfo:
        return self._get_pos_economics(block_identifier)
    
    def get_pos_reward_by_epoch(
        self, epoch_number: Union[EpochNumber, int]
    ) -> Union[PoSEpochRewardInfo, None]:
        return self._get_pos_reward_by_epoch(epoch_number)
    
    def get_params_from_vote(
        self, block_identifier: Optional[EpochNumberParam] = None
    ) -> DAOVoteInfo:
        return self._get_params_from_vote(block_identifier)
    
    def get_supply_info(self) -> SupplyInfo:
        return self._get_supply_info()
    
    def get_account_pending_info(
        self, address: Union[Base32Address, str]
    ) -> PendingInfo:
        return self._get_account_pending_info(address)
    
    def get_account_pending_transactions(
        self, address: Union[Base32Address, str], start_nonce: Optional[int]=None, limit: Optional[int]=None
    ) -> PendingTransactionsInfo:
        return self._get_account_pending_transactions(address, start_nonce, limit)
    
    def check_balance_against_transaction(
        self,
        account_address: Union[Base32Address, str],
        contract_address: Union[Base32Address, str], 
        gas_limit: int,
        gas_price: Union[Drip, AbstractDerivedTokenUnit[Drip], int],
        storage_limit: Union[Storage, int],
        block_identifier: Optional[EpochNumberParam] = None
    ) -> TransactionPaymentInfo:
        gas_price = to_int_if_drip_units(gas_price)
        return self._check_balance_against_transaction(
            account_address, contract_address, gas_limit, gas_price, storage_limit, block_identifier
        )
    
    @overload
    def get_logs(self, filter_params: FilterParams) -> List[LogReceipt]:...
    
    @overload
    def get_logs(self, filter_params: None=None, **kwargs: Any) -> List[LogReceipt]:...
    
    def get_logs(self, filter_params: Optional[FilterParams]=None, **kwargs: Any) -> List[LogReceipt]:
        """
        Returns logs matching the filter provided.
        It is accepted to pass filter_params as a dict or by direclty specifying field name (but cannot mix)

        >>> logs = w3.cfx.get_logs({"fromEpoch": 97134060, "toEpoch: 97134160"})
        >>> assert logs == w3.cfx.get_logs(fromEpoch=97134060, toEpoch=97134160})

        Parameters
        ----------
        filter_params : Optional[FilterParams], optional
            visit https://developer.confluxnetwork.org/conflux-doc/docs/json_rpc/#cfx_getlogs for more information

        Returns
        -------
        List[LogReceipt]
            a list of LogReceipt. It is recommended to read https://github.com/Conflux-Chain/python-conflux-sdk/blob/v1/examples/04-interact_with_contracts_and_logs.py to know how to process the returned logs
        """        
        if filter_params is None:
            filter_params = cast(FilterParams, keyfilter(lambda key: key in FilterParams.__annotations__.keys(), kwargs)) # type: ignore
            return self._get_logs(filter_params)
        else:
            if len(kwargs.keys()) != 0:
                raise ValueError("Redundant Param: FilterParams as get_logs first parameter is already provided")
            return self._get_logs(filter_params)

    def get_collateral_info(self, block_identifier: Optional[EpochNumberParam] = None) -> CollateralInfo:
        return self._get_collateral_info(block_identifier)

    @overload
    def new_filter(self, filter_params: FilterParams) -> LogFilterId:...
    
    @overload
    def new_filter(self, filter_params: None=None, **kwargs: Unpack[FilterParams]) -> LogFilterId:...
    
    def new_filter(self, filter_params: Optional[FilterParams]=None, **kwargs: Unpack[FilterParams]) -> LogFilterId:     
        if filter_params is None:
            filter_params = cast(FilterParams, keyfilter(lambda key: key in FilterParams.__annotations__.keys(), kwargs)) # type: ignore
            return self._new_filter(filter_params)
        else:
            if len(kwargs.keys()) != 0:
                raise ValueError("Redundant Param: FilterParams as get_logs first parameter is already provided")
            return self._new_filter(filter_params)
        
    def new_block_filter(self) -> BlockFilterId:
        return self._new_block_filter()
    
    def new_pending_transaction_filter(self) -> TxFilterId:
        return self._new_pending_transaction_filter()
    
    
    @overload
    def get_filter_changes(self, filter_id: LogFilterId) -> Sequence[LogReceipt]:...
    
    @overload
    def get_filter_changes(self, filter_id: Union[TxFilterId, BlockFilterId]) -> Sequence[Hash32]:...
    
    @overload
    def get_filter_changes(self, filter_id: str) -> Union[Sequence[Hash32], Sequence[LogReceipt]]:...
    
    
    def get_filter_changes(self, filter_id: _FilterId) -> Union[Sequence[Hash32], Sequence[LogReceipt]]:     
        return self._get_filter_changes(filter_id)

    def get_filter_logs(self, log_filter_id: Union[LogFilterId, str]) -> Sequence[LogReceipt]:
        return self._get_filter_logs(log_filter_id)

    def uninstall_filter(self, filter_id: _FilterId) -> bool:
        return self._uninstall_filter(filter_id)
    
    def get_admin(self, address: AddressParam, block_identifier: Optional[EpochNumberParam] = None) -> Union[None, Base32Address]:
        return self._get_admin(address, block_identifier)
    
    def fee_history(self, epoch_count: int, block_identifier: EpochNumberParam, reward_percentiles: Sequence[float]) -> FeeHistory:
        return self._fee_history(epoch_count, block_identifier, reward_percentiles)
    
    def get_fee_burnt(self) -> Drip:
        return self._get_fee_burnt()
    
    @property
    def max_priority_fee(self) -> Drip:
        return self._max_priority_fee()
    
    def set_gas_price_strategy(self, strategy: Callable[["Web3", TxParam], Drip]):
        super().set_gas_price_strategy(strategy) # type: ignore



from typing import TYPE_CHECKING
from conflux_web3._utils.rpc_abi import (
    RPC
)
from conflux_web3.types.transaction_hash import (
    TransactionHash
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

    
class PendingTransactionMiddleware:
    def __init__(self, make_request, w3: "Web3"):
        self._make_request = make_request
        self._w3 = w3
        
    def __call__(self, method, params):
        response = self._make_request(method, params)
        if method == RPC.cfx_sendTransaction or method == RPC.cfx_sendRawTransaction:
            if "result" in response:
                transaction_hash = TransactionHash(response["result"])
                transaction_hash.set_w3(self._w3)
                
                response["result"] = transaction_hash
        return response




from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    NewType,
    Optional,
    Sequence,
    Union,
    Dict
)
from typing_extensions import (
    Literal,
    TypedDict,
)
from hexbytes import HexBytes

from web3.types import (
    RPCEndpoint,
    RPCResponse,
)
from web3.datastructures import (
    NamedElementOnion,
)

from cfx_address import Base32Address
from cfx_utils.types import (
    TxDict,
    TxParam,
    HexAddress,
    Hash32, # Hash32 as specific output data type
    _Hash32, # _Hash32 as robust input data type
    Nonce,
    Drip,
    CFX,
    GDrip,
    AddressParam,
    Storage,
    EpochNumberParam,
    EpochLiteral,
    EpochNumber,
    HexStr,
    BlockNumber,
    CIP1559TxDict,
    LegacyTxDict,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

class NodeStatus(TypedDict):
    """
    Node status as dict

    Parameters
    ----------
    | bestHash1111: bytes
    | chainId: int
    | networkId: int
    | blockNumber: int
    | epochNumber: int
    | latestCheckpoint: int
    | latestConfirmed: int
    | latestState: int
    | latestFinalized: int
    | ethereumSpaceChainId: int
    | pendingTxNumber: int
    """    
    bestHash: Hash32
    chainId: int
    networkId: int
    blockNumber: int
    epochNumber: int
    latestCheckpoint: int
    latestConfirmed: int
    latestState: int
    latestFinalized: int
    ethereumSpaceChainId: int
    pendingTxNumber: int

class EstimateResult(TypedDict):
    """
    Estimation result of a transaction as dict

    Parameters
    ----------
    | gasLimit: int
    | gasUsed: int
    | storageCollateralized: Storage
    """       
    gasLimit: int
    gasUsed: int
    storageCollateralized: Storage


class FilterParams(TypedDict, total=False):
    """
    Parameter dict to filter logs, more information at https://developer.confluxnetwork.org/conflux-doc/docs/json_rpc/#cfx_getlogs

    Parameters
    ----------
    | fromEpoch: EpochNumberParam, optional
    | toEpoch: EpochNumberParam, optional
    | fromBlock: int, optional
    | toBlock: int, optional
    | blockHashes: Sequence[_Hash32], optional
    | address: Union[Base32Address, List[Base32Address]], optional
    | topics: Sequence[Optional[Union[_Hash32, Sequence[_Hash32]]]], optional
    """
    fromEpoch: EpochNumberParam
    toEpoch: EpochNumberParam
    fromBlock: BlockNumber
    toBlock: BlockNumber
    blockHashes: Sequence[_Hash32]
    address: Union[Base32Address, List[Base32Address]]
    topics: Sequence[Optional[Union[_Hash32, Sequence[_Hash32]]]]


class TransactionLogReceipt(TypedDict):
    """
    Log receipt in transaction receipt

    Parameters
    ----------
    | address: Base32Address
    | topics: Sequence[HexBytes]
    | data: HexBytes
    """
    address: Base32Address
    topics: Sequence[HexBytes]
    data: HexBytes


class LogReceipt(TypedDict):
    """
    Full log receipt

    Parameters
    ----------
    | address: Base32Address
    | topics: Sequence[HexBytes]
    | data: HexBytes
    | blockHash: Hash32
    | epochNumber: int
    | transactionHash: Hash32
    | transactionIndex: int
    | logIndex: int
    | transactionLogIndex: int
    """
    address: Base32Address
    topics: Sequence[HexBytes]
    data: HexBytes
    blockHash: Hash32
    epochNumber: int
    transactionHash: Hash32
    transactionIndex: int
    logIndex: int
    transactionLogIndex: int

class TransactionEventData(TypedDict):
    """
    Transaction event data as a dict

    Parameters
    ----------
    | address: Base32Address
    | args: Dict[str, Any]
    | event: str
    | blockHash: Hash32
    | epochNumber: int
    | transactionHash: Hash32
    | transactionIndex: int
    | transactionLogIndex: int
    """
    address: Base32Address
    args: Dict[str, Any]
    event: str
    blockHash: Hash32
    epochNumber: int
    transactionHash: Hash32
    transactionIndex: int
    transactionLogIndex: int

class EventData(TransactionEventData, total=False):
    """
    Transaction event data

    Parameters
    ----------
    | address: Base32Address
    | args: Dict[str, Any]
    | event: str
    | blockHash: Hash32
    | epochNumber: int
    | transactionHash: Hash32
    | transactionIndex: int
    | transactionLogIndex: int
    | logIndex: int, Optional
    """
    logIndex: int


# syntax b/c "from" keyword not allowed w/ class construction
TxReceipt = TypedDict(
    "TxReceipt",
    {
        "type": int,
        "transactionHash": Hash32,
        "index": int,
        "blockHash": Hash32,
        "epochNumber": int,
        "from": Base32Address,
        "to": Union[Base32Address, None],
        "gasUsed": int,
        "gasFee": Drip,
        "gasCoveredBySponsor": bool,
        "storageCollateralized": Storage,
        "storageCoveredBySponsor": bool,
        "storageReleased": List[Storage],
        "contractCreated": Union[Base32Address, None],
        "stateRoot": Hash32,
        "outcomeStatus": int,
        "logsBloom": HexBytes,
        "logs": List[TransactionLogReceipt],
        "txExecErrorMsg": Union[str, None],
        "effectiveGasPrice": Drip,
        "burntGasFee": Drip,
    },
)
"""
Transaction receipt as a dict

Parameters
----------
| "type": int,
| "transactionHash": Hash32,
| "index": int,
| "blockHash": Hash32,
| "epochNumber": int,
| "from": Base32Address,
| "to": Base32Address,
| "gasUsed": int,
| "gasFee": Drip,
| "gasCoveredBySponsor": bool,
| "storageCollateralized": Storage,
| "storageCoveredBySponsor": bool,
| "storageReleased": List[Storage],
| "contractCreated": Union[Base32Address, None],
| "stateRoot": Hash32,
| "outcomeStatus": int,
| "logsBloom": HexBytes,
| "logs": List[TransactionLogReceipt],
| "txExecErrorMsg": Union[str, None]
| "effectiveGasPrice": Drip,
| "burntGasFee": Drip,
"""

class TxReceiptWithSpace(TxReceipt):
    space: Literal["native", "evm"]

class AccessListEntry(TypedDict):
    address: Base32Address
    storageKeys: Sequence[HexStr]

# syntax b/c "from" keyword not allowed w/ class construction
LegacyTxData = TypedDict(
    "LegacyTxData",
    {
        "type": int,
        "blockHash": Union[None, Hash32],
        "chainId": int,
        "contractCreated": Union[None, Base32Address],
        "data": HexBytes,
        "epochHeight": int,
        "from": Base32Address,
        "gas": int,
        "gasPrice": Drip,
        "hash": Hash32,
        "nonce": Nonce,
        "r": HexBytes,
        "s": HexBytes,
        "status": Union[None, int],
        "storageLimit": Storage,
        "to": Union[None, Base32Address],
        "transactionIndex": Union[None, int],
        "v": int,
        "value": Drip,
    },
    total=False,
)

class TxData(LegacyTxData, total=False):
    """
    Transaction data as a dict

    Parameters
    ----------
    | "type": int,
    | "blockHash": Union[None, Hash32],
    | "chainId": int,
    | "contractCreated": Union[None, Base32Address],
    | "data": HexBytes,
    | "epochHeight": int,
    | "from": Base32Address,
    | "gas": int,
    | "gasPrice": Drip,
    | "hash": Hash32,
    | "nonce": Nonce,
    | "r": HexBytes,
    | "s": HexBytes,
    | "status": Union[None, int],
    | "storageLimit": Storage,
    | "to": Union[None, Base32Address],
    | "transactionIndex": Union[None, int],
    | "v": int,
    | "value": Drip,
    """
    maxFeePerGas: Drip
    maxPriorityFeePerGas: Drip
    accessList:  Sequence[AccessListEntry]
    yParity: int

class BlockData(TypedDict):
    """
    Block data as a dict

    Parameters
    ----------
    | hash: Hash32
    | parentHash: Hash32
    | height: int
    | miner: Base32Address
    | deferredStateRoot: Hash32
    | deferredReceiptsRoot: Hash32
    | deferredLogsBloomHash: Hash32
    | blame: int
    | transactionsRoot: Hash32
    | epochNumber: Union[int, None]
    | blockNumber: Union[int, None]
    | gasLimit: int
    | gasUsed: Union[int, None]
    | timestamp: int
    | difficulty: int
    | powQuality: Union[HexBytes, None]
    | refereeHashes: Sequence[Hash32]
    | adaptive: bool
    | nonce: HexBytes
    | size: int
    | custom: Sequence[HexBytes]
    | posReference: Hash32
    | transactions: Sequence[Union[Hash32, TxData]]
    """
    hash: Hash32
    parentHash: Hash32
    height: int
    miner: Base32Address
    deferredStateRoot: Hash32
    deferredReceiptsRoot: Hash32
    deferredLogsBloomHash: Hash32
    blame: int
    transactionsRoot: Hash32
    epochNumber: Union[int, None]
    blockNumber: Union[int, None]
    gasLimit: int
    gasUsed: Union[int, None]
    timestamp: int
    difficulty: int
    powQuality: Union[HexBytes, None]
    refereeHashes: Sequence[Hash32]
    adaptive: bool
    nonce: HexBytes # block nonce put by miner rather than the nonce in the transaction
    size: int
    custom: Sequence[HexBytes]
    posReference: Hash32
    transactions: Sequence[Union[Hash32, TxData]]
    baseFeePerGas: Drip
    

Middleware = Callable[[Callable[[RPCEndpoint, Any], RPCResponse], "Web3"], Any]
MiddlewareOnion = NamedElementOnion[str, Middleware]

class StorageRoot(TypedDict):
    """

    Parameters
    ----------
    | delta: Union[Hash32, Literal["TOMBSTONE", None]]
    | intermediate: Union[Hash32, Literal["TOMBSTONE", None]]
    | snapshot: Union[Hash32, Literal["TOMBSTONE", None]]
    """
    delta: Union[Hash32, Literal["TOMBSTONE", None]]
    intermediate: Union[Hash32, Literal["TOMBSTONE", None]]
    snapshot: Union[Hash32, Literal["TOMBSTONE", None]]

class SponsorInfo(TypedDict):
    """

    Parameters
    ----------
    | sponsorBalanceForCollateral: Drip
    | sponsorBalanceForGas: Drip
    | sponsorForCollateral: Base32Address
    | sponsorForGas: Base32Address
    | sponsorGasBound: Drip
    | usedStoragePoints: int
    | availableStoragePoints: int
    """
    sponsorBalanceForCollateral: Drip
    sponsorBalanceForGas: Drip
    sponsorForCollateral: Base32Address
    sponsorForGas: Base32Address
    sponsorGasBound: Drip
    usedStoragePoints: int
    availableStoragePoints: int
    
class AccountInfo(TypedDict):
    """

    Parameters
    ----------
    | address: Base32Address
    | balance: Drip
    | nonce: Nonce
    | codeHash: Hash32
    | stakingBalance: Drip
    | collateralForStorage: Storage
    | accumulatedInterestReturn: Drip
    | admin: Base32Address
    """
    address: Base32Address
    balance: Drip
    nonce: Nonce
    codeHash: Hash32
    stakingBalance: Drip
    collateralForStorage: Storage
    accumulatedInterestReturn: Drip
    admin: Base32Address
    
class DepositInfo(TypedDict):
    """

    Parameters
    ----------
    | accumulatedInterestRate: int
    | amount: Drip
    | depositTime: int
    """
    accumulatedInterestRate: int
    amount: Drip
    depositTime: int # assumed blockNumber
    
class VoteInfo(TypedDict):
    """

    Parameters
    ----------
    | amount: Drip
    | unlockBlockNumber: int
    """
    amount: Drip
    unlockBlockNumber: int

class BlockRewardInfo(TypedDict):
    """

    Parameters
    ----------
    | blockHash: Hash32
    | author: Base32Address
    | totalReward: Drip
    | baseReward: Drip
    | txFee: Drip
    """
    blockHash: Hash32
    author: Base32Address
    totalReward: Drip
    baseReward: Drip
    txFee: Drip

PoSBlockNumber = NewType("PoSBlockNumber", int)
PoSEpochNumber = NewType("PoSEpochNumber", int)

class PoSEconomicsInfo(TypedDict):
    """

    Parameters
    ----------
    | distributablePosInterest: Drip
    | lastDistributeBlock: PoSBlockNumber
    | totalPosStakingTokens: Drip
    """
    distributablePosInterest: Drip
    lastDistributeBlock: PoSBlockNumber
    totalPosStakingTokens: Drip
    
class PoSAccountRewardsInfo(TypedDict):
    """

    Parameters
    ----------
    | posAddress: Base32Address
    | powAddress: Base32Address
    | reward: Drip
    """
    posAddress: Base32Address
    powAddress: Base32Address
    reward: Drip

class PoSEpochRewardInfo(TypedDict):
    """

    Parameters
    ----------
    | accountRewards: Sequence[PoSAccountRewardsInfo]
    | powEpochHash: Hash32
    """
    accountRewards: Sequence[PoSAccountRewardsInfo]
    powEpochHash: Hash32

class DAOVoteInfo(TypedDict):
    """

    Parameters
    ----------
    | powBaseReward: Drip
    | interestRate: int
    | storagePointProp: int
    | baseFeeShareProp: int
    """
    powBaseReward: Drip
    interestRate: int
    storagePointProp: int
    baseFeeShareProp: int

class SupplyInfo(TypedDict):
    """

    Parameters
    ----------
    | totalIssued: Drip
    | totalCollateral: Drip
    | totalStaking: Drip
    | totalCirculating: Drip
    | totalEspaceTokens: Drip
    """
    totalIssued: Drip
    totalCollateral: Drip
    totalStaking: Drip
    totalCirculating: Drip
    totalEspaceTokens: Drip

class PendingInfo(TypedDict):
    """

    Parameters
    ----------
    | localNonce: Nonce
    | pendingNonce: Nonce
    | pendingCount: int
    | nextPendingTx: Hash32
    """
    localNonce: Nonce
    pendingNonce: Nonce
    pendingCount: int
    nextPendingTx: Hash32

class PendingTransactionStatus(TypedDict):
    pending: Literal["futureNonce", "notEnoughCash"]

class PendingTransactionsInfo(TypedDict):
    """

    Parameters
    ----------
    | firstTxStatus: Union[PendingTransactionStatus, Literal["ready"]]
    | pendingCount: int
    | pendingTransactions: Sequence[TxData]
    """
    firstTxStatus: Union[PendingTransactionStatus, Literal["ready"]]
    pendingCount: int
    pendingTransactions: Sequence[TxData]

class TransactionPaymentInfo(TypedDict):
    """

    Parameters
    ----------
    | isBalanceEnough: bool
    | willPayCollateral: bool
    | willPayTxFee: bool
    """
    isBalanceEnough: bool
    willPayCollateral: bool
    willPayTxFee: bool

class CollateralInfo(TypedDict):
    """

    Parameters
    ----------
    | totalStorageTokens: int
    | convertedStoragePoints: int
    | usedStoragePoints: int
    """
    totalStorageTokens: int
    convertedStoragePoints: int
    usedStoragePoints: int
    
LogFilterId = NewType("LogFilterId", HexStr)
BlockFilterId = NewType("BlockFilterId", HexStr)
TxFilterId = NewType("TxFilterId", HexStr)

_FilterId = Union[LogFilterId, BlockFilterId, TxFilterId, str]

class FeeHistory(TypedDict):
    """
    Fee history information

    Parameters
    ----------
    | baseFeePerGas: Sequence[Drip] - List of base fee per gas values for each epoch
    | gasUsedRatio: Sequence[float] - List of gas used ratios for each epoch pivot block
    | oldestEpoch: int - The oldest epoch number in the returned range
    | reward: Sequence[Sequence[Drip]] - List of effective priority fee per gas values for each epoch
    """
    baseFeePerGas: Sequence[Drip]
    gasUsedRatio: Sequence[float]
    oldestEpoch: int
    reward: Sequence[Sequence[Drip]]



from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Union,
    Callable,
    Any,
)
import warnings

from eth_keys.datatypes import (
    PrivateKey,
)

from cfx_utils.types import (
    HexAddress,
    TxDict,
)
from cfx_account.account import (
    LocalAccount,
    Account,
)
from cfx_address import (
    Base32Address
)
from cfx_address.utils import (
    validate_network_id as validate_chain_id,
    normalize_to
)
from conflux_web3._utils.cns import (
    resolve_if_cns_name,
)
from conflux_web3._utils.rpc_abi import (
    RPC
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

_PrivateKey = Union[LocalAccount, PrivateKey, str, bytes]


class Wallet:
    def __init__(self, 
                account_or_accounts: Union[Iterable[_PrivateKey], _PrivateKey]=[],
                forced_chain_id: Optional[int]=None,
                ):
        """
        generate a wallet middleware object with specific chain_id and accounts to use.
        Note: this CLASS is not a web3.middleware. 
        A Wallet-type INSTANCE is the actual middleware

        Parameters
        ----------
        account_or_accounts : Union[Sequence[_PrivateKey], _PrivateKey], optional, by default []
            Any param could be private key source. 
            Both [account] and account can be served as "accounts" param.
            For LocalAccount type param, ensure chain_id and LocalAccount is consistent. 
        forced_chain_id : Optional[int], optional, by default None
            The network id of the wallet, all account will be set at the specified network, 
            and network checking will be applied to every added account. 
            If is None, then no default network id is set, and this wallet can be used in any network.
        """
        if forced_chain_id is not None:
            validate_chain_id(forced_chain_id)

        self._chain_id = forced_chain_id
        self._accounts_map: Dict[str, LocalAccount] = {}
        
        if isinstance(account_or_accounts, Iterable):
            for account in account_or_accounts:
                self.add_account(account)
        else:
            self.add_account(account_or_accounts)
    
    @property
    def chain_id(self):
        return self._chain_id
    
    @property
    def forced_chain_id(self):
        return self._chain_id
    
    @forced_chain_id.setter
    def forced_chain_id(self, new_chain_id: Union[None, int]):
        """
        the forced chain id of the wallet.
        if set to not None, all account in the wallet will be converted to the corresponding network.
        After that, accounts from incompatible network cannot be added to this wallet,
        and signing requests from network other than forced_chain_id will be ignored 

        Parameters
        ----------
        new_chain_id : Union[None, int] 
            _description_
        """        
        self._chain_id = new_chain_id
        for old_address, account in self._accounts_map.copy().items():
            account.network_id = new_chain_id
            self._accounts_map.pop(old_address)
            self.add_account(account)
        self._chain_id = new_chain_id
    
    @property
    def accounts(self) -> Sequence[Union[Base32Address, HexAddress]]:
        """
        returns all accounts address in the wallet

        Returns
        -------
        Sequence[Union[Base32Address, HexAddress]]
            a sequence of Base32Address if wallet.forced_chain_id is not None
            or HexAddress if wallet.forced_chain_id is None
        """            
        return self._accounts_map.keys() # type: ignore
    
    
    def normalize_private_key_to_account(self, private_key: _PrivateKey) -> LocalAccount:
        if isinstance(private_key, LocalAccount):
            local_account = private_key
            # if self._chain_id is set, check if local_account.network_id is same as self._network_id
            if self._chain_id:
                if local_account.network_id and local_account.network_id != self._chain_id:
                    raise ValueError("wallet's chain_id and local_account's chain_id is supposed to be consistent")
            private_key = local_account.key
        # any account added to wallet is a brand new object
        return Account.from_key(private_key, self._chain_id)
    
    def __call__(self, make_request: Callable[..., Dict[str, Any]], w3: "Web3"):
        def inner(method: str, params: Sequence[Any]):
            if method != RPC.cfx_sendTransaction:
                return make_request(method, params)
            
            transaction: TxDict = params[0]
            if "from" not in transaction:
                return make_request(method, params)
            else:
                transaction["from"] = resolve_if_cns_name(w3, transaction["from"])
                if transaction["from"] not in self:
                    return make_request(method, params)
            
            account = self[transaction["from"]]
            raw_tx = account.sign_transaction(transaction).rawTransaction
            # because param formatting has been done before middleware process
            # we do the param formatting manually
            response = make_request(RPC.cfx_sendRawTransaction, [raw_tx.hex()])
            return response
        return inner

    def add_account(self, account: _PrivateKey):
        local_account = self.normalize_private_key_to_account(account)
        # assert Address.has_network_prefix
        if local_account.address in self._accounts_map:
            warnings.warn(f"Duplicate account: {local_account.address} is already in the wallet, this operation overwrites the existed old account")
        self._accounts_map[local_account.address] = local_account

    def add_accounts(self, accounts: Iterable[_PrivateKey]):
        for account in accounts:
            self.add_account(account)
    
    def __getitem__(self, address: str) -> LocalAccount:
        if self._chain_id is None:
            return self._accounts_map[normalize_to(address, None)]
        else:
            return self._accounts_map[address]
    
    def __contains__(self, address: str) -> bool:
        try:
            self[address]
            return True
        except KeyError:
            return False
    
    def pop(self, address: str) -> LocalAccount:
        if self._chain_id is None:
            return self._accounts_map.pop(normalize_to(address, None))
        else:
            return self._accounts_map.pop(address)


def construct_sign_and_send_raw_middleware(
    account_or_accounts: Union[Sequence[_PrivateKey], _PrivateKey], 
    forced_chain_id: Optional[int]=None
) -> Wallet:
    return Wallet(account_or_accounts, forced_chain_id)



from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    cast,
    List,
    Callable,
    Union,
)

from web3.contract import (
    Contract,
)

from web3.contract.utils import (
    find_functions_by_identifier,
)

from web3._utils.normalizers import (
    normalize_abi,
    normalize_bytecode,
)
from web3._utils.datatypes import (
    PropertyCheckingFactory,
)
from web3.types import (
    ABI
)

from cfx_address import (
    Base32Address,
)
from cfx_address.utils import (
    validate_address_agaist_network_id,
)

from cfx_utils.exceptions import (
    InvalidEpochNumebrParam,
    Base32AddressNotMatch,
)
from cfx_utils.decorators import (
    combomethod,
)
from cfx_utils.types import (
    ChecksumAddress,
)
from conflux_web3.types import (
    Base32Address,
    AddressParam,
    EpochNumberParam,
    EpochLiteral,
)
from conflux_web3._utils.cns import (
    resolve_if_cns_name
)
from conflux_web3.contract.function import (
    ConfluxContractFunction,
    ConfluxContractFunctions,
)
from conflux_web3.contract.caller import (
    ConfluxContractCaller
)
from conflux_web3.contract.event import (
    ConfluxContractEvents
)
from conflux_web3.contract.constructor import (
    ConfluxContractConstructor
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

# # used to hook web3.contract.parse_block_identifier
# # hook is activated in _web3_hook
# def cfx_parse_block_identifier(
#     w3: "Web3", block_identifier: EpochNumberParam
# ) -> EpochNumberParam:
#     if isinstance(block_identifier, int):
#         return block_identifier
#     elif block_identifier in EpochLiteral.__args__: # type: ignore
#         return block_identifier
#     elif isinstance(block_identifier, bytes) or is_hex_encoded_block_hash(
#         block_identifier
#     ):
#         # r = 
#         # assert r is not None
#         return w3.cfx.get_block_by_hash(block_identifier)["epochNumber"] # type: ignore
#     else:
#         raise InvalidEpochNumebrParam


class ConfluxContract(Contract):
    address: Base32Address
    w3: 'Web3'
    functions: ConfluxContractFunctions
    caller: "ConfluxContractCaller"
    events: "ConfluxContractEvents"
    
    def __init__(self, address: AddressParam) -> None:
        """Create a new smart contract proxy object.

        :param address: Base32 Contract address 
        """
        if self.w3 is None:
            raise AttributeError(
                'The `Contract` class has not been initialized.  Please use the '
                '`web3.contract` interface to create your contract class.'
            )

        # address should match chainId
        if address:
            address = resolve_if_cns_name(self.w3, address)
            validate_address_agaist_network_id(address, self.w3.cfx.chain_id, True)
            address = Base32Address(address, self.w3.cfx.chain_id)
            if address.address_type != "contract" and address.address_type != "builtin":
                raise Base32AddressNotMatch(f"expected an address of contract type or builtin type"
                                            f"receives {address} of {address.address_type}")
            self.address = Base32Address(address, self.w3.cfx.chain_id, verbose=True)

        if not self.address:
            raise TypeError("The address argument is required to instantiate a contract.")

        self.functions = ConfluxContractFunctions(self.abi, self.w3, self.address, decode_tuples=self.decode_tuples)
        self.caller = ConfluxContractCaller(self.abi, self.w3, self.address, decode_tuples=self.decode_tuples)
        self.events = ConfluxContractEvents(self.abi, self.w3, self.address)
        self.fallback = Contract.get_fallback_function(self.abi, self.w3, ConfluxContractFunction, self.address) # type: ignore
        self.receive = Contract.get_receive_function(self.abi, self.w3, ConfluxContractFunction, self.address) # type: ignore

    @classmethod
    def factory(cls, w3: "Web3", class_name: Optional[str] = None, **kwargs: Any) -> "Contract":
        kwargs["w3"] = w3

        normalizers = {
            "abi": normalize_abi,
            # "address": partial(normalize_address, kwargs["w3"].ens),
            "bytecode": normalize_bytecode,
            "bytecode_runtime": normalize_bytecode,
        }

        contract = cast(
            ConfluxContract,
            PropertyCheckingFactory(
                class_name or cls.__name__,
                (cls,),
                kwargs,
                normalizers=normalizers,
            ),
        )
        contract.functions = ConfluxContractFunctions(contract.abi, contract.w3, decode_tuples=contract.decode_tuples)
        contract.caller = ConfluxContractCaller(contract.abi, contract.w3, contract.address, decode_tuples=contract.decode_tuples)
        contract.events = ConfluxContractEvents(contract.abi, contract.w3)
        contract.fallback = Contract.get_fallback_function(
            contract.abi,
            contract.w3,
            ConfluxContractFunction,
        )
        contract.receive = Contract.get_receive_function(
            contract.abi,
            contract.w3,
            ConfluxContractFunction,
        )

        return contract

    @classmethod
    def constructor(cls, *args: Any, **kwargs: Any) -> "ConfluxContractConstructor":
        """
        :param args: The contract constructor arguments as positional arguments
        :param kwargs: The contract constructor arguments as keyword arguments
        :return: a contract constructor object
        """
        if cls.bytecode is None:
            raise ValueError(
                "Cannot call constructor on a contract that does not have "
                "'bytecode' associated with it"
            )

        return ConfluxContractConstructor(cls.w3, cls.abi, cls.bytecode, *args, **kwargs)

    @combomethod
    def find_functions_by_identifier(
        cls,
        contract_abi: ABI,
        w3: "Web3",
        address: Union[ChecksumAddress, Base32Address],
        callable_check: Callable[..., Any],
    ) -> List["ConfluxContractFunction"]:
        return find_functions_by_identifier(  # type: ignore
            contract_abi, w3, address, callable_check, ConfluxContractFunction # type: ignore
        )



from typing import (
    Any,
    ForwardRef,
    Type,
    Union,
    cast,
)
import typing
import sys
from typing_extensions import (
    Literal,
    TypedDict, 
    get_args,
    get_origin, 
    is_typeddict,
)
import collections.abc

import conflux_web3.types

class TypeValidator:
    """
    use TypedDict.__annotations__ to check type

    >> from typing import Union, get_origin, get_args
    >> x = Union[int, str]
    >> get_origin(x), get_args(x)
    (typing.Union, (<class 'int'>, <class 'str'>))
    >> get_origin(x) is Union
    True
    >> get_args(int)
    ()
    >> isinstance(3, get_args(x))
    True
    >> isinstance('a', get_args(x))
    True
    >> isinstance([], get_args(x))
    False
    """
    @staticmethod
    def _is_list_like_type(typ: type):
        return any(
            [
                get_origin(typ) == collections.abc.Sequence,
                get_origin(typ) == list
            ]
        )
    
    @staticmethod
    def isinstance(val: Any, field_type: Type[Any]) -> bool:
        if is_typeddict(field_type):
            annotations = field_type.__annotations__
            for key, sub_field_type in annotations.items():
                if key not in val:
                    return False
                if not TypeValidator.isinstance(val[key], sub_field_type):
                    return False
            return True
        if get_origin(field_type) is Union:
            return any(TypeValidator.isinstance(val, t)
                        for t in get_args(field_type))
        if get_origin(field_type) is Literal:
            return val in get_args(field_type)
        if TypeValidator._is_list_like_type(field_type):
            return all(
                TypeValidator.isinstance(v, get_args(field_type)[0])
                for v in val
            )
        if sys.version_info >= (3, 10):
            # NewType becomes a class after python3.10
            if type(field_type) == typing.NewType:
                return isinstance(val, field_type.__supertype__)
        elif type(field_type).__name__ == "function":
            return isinstance(val, field_type.__supertype__) # type: ignore
        if type(field_type) is type:
            # for sake of debug
            if isinstance(val, field_type):
                return True
            return False
        if type(field_type) is ForwardRef:
            return type(val).__name__ == field_type.__forward_arg__
        if type(field_type) is typing._GenericAlias: # type: ignore
            return TypeValidator.isinstance(val, cast(type, get_origin(field_type)))
        else:
            if isinstance(val, field_type):
                return True
            return False
            # warnings.warn("complex type check")
            # return True
        
    @staticmethod
    def assert_instance(val: Any, field_type: Type[Any]) -> Literal[True]:
        from pydantic import BaseModel, ValidationError

        if is_typeddict(field_type):
            class PydanticModel(BaseModel, arbitrary_types_allowed=True):
                __annotations__ = field_type.__annotations__

            try:
                PydanticModel(**val)
            except ValidationError as e:
                raise TypeError(f"Validation error for typed_dict: {e}")
            return True
        if get_origin(field_type) is Union:
            for t in get_args(field_type):
                if TypeValidator.isinstance(val, t):
                    return True
            raise TypeError(f"value does not match union: {val} does not match any of {get_args(field_type)}")
        if get_origin(field_type) is Literal:
            if val in get_args(field_type):
                return True
            raise TypeError(f"value does not match literal: {val} is not any of {get_args(field_type)}")
        if TypeValidator._is_list_like_type(field_type):
            for v in val:
                try:
                    TypeValidator.assert_instance(v, get_args(field_type)[0])
                except TypeError as e:
                    raise TypeError(f"some value in list are not in expecetd type: {get_args(field_type)[0]}") from e
        if sys.version_info >= (3, 10):
            # NewType becomes a class after python3.10
            if type(field_type) == typing.NewType:
                return isinstance(val, field_type.__supertype__)
        elif type(field_type).__name__ == "function":
            if isinstance(val, field_type.__supertype__): # type: ignore
                return True
            raise TypeError(f"value does not match type: {val} is not type {field_type.__supertype__}") # type: ignore
        if type(field_type) is type:
            # for sake of debug
            if isinstance(val, field_type):
                return True
            raise TypeError(f"value does not match type: {val} is not type {field_type}")
        if type(field_type) is ForwardRef:
            return type(val).__name__ == field_type.__forward_arg__
        if type(field_type) is typing._GenericAlias: # type: ignore
            return TypeValidator.assert_instance(val, cast(type, get_origin(field_type)))
        else:
            if isinstance(val, field_type):
                return True
            raise TypeError(f"value does not match type: {val} is not type {field_type}")

    @staticmethod
    def validate_typed_dict(value_to_validate: Any, typed_dict_class: Union[str, Type[TypedDict]]):
        if isinstance(typed_dict_class, str):
            typed_dict_class = cast(Type[TypedDict], getattr(conflux_web3.types, typed_dict_class))
        TypeValidator.assert_instance(value_to_validate, typed_dict_class)
        return True
    
    @staticmethod
    def validate_tx_data(value_to_validate: conflux_web3.types.TxData):
        val = dict(value_to_validate)
        if "maxFeePerGas" not in val:
            val["maxFeePerGas"] = conflux_web3.types.Drip(0)
        if "maxPriorityFeePerGas" not in value_to_validate:
            val["maxPriorityFeePerGas"] = conflux_web3.types.Drip(0)
        if "accessList" not in val:
            val["accessList"] = []
        if "yParity" not in val:
            val["yParity"] = 0
        return TypeValidator.validate_typed_dict(val, "TxData")
        


from typing import Any
from eth_abi.registry import (
    ABIRegistry,
    BaseEquals
)
from eth_abi.decoding import (
    AddressDecoder
)
from eth_abi.exceptions import (
    EncodingError
)
from web3._utils.abi import (
    build_non_strict_registry, 
    AddressEncoder,
)
from cfx_address import (
    Base32Address,
    validate_base32
)
from cfx_utils.exceptions import (
    InvalidBase32Address
)
from conflux_web3._utils.cns import (
    is_cns_name
)

class Base32AddressEncoder(AddressEncoder):
    
    encode_fn = lambda self, address: AddressEncoder.encode_fn(Base32Address(address).hex_address)
    
    @classmethod
    def validate_value(cls, value: Any) -> None:
        if is_cns_name(value):
            return
        try:
            validate_base32(value)
        except InvalidBase32Address:
            raise EncodingError(InvalidBase32Address)

class CfxAddressDecoder(AddressDecoder):
    decode_fn = lambda x: x

def build_cfx_default_registry() -> ABIRegistry:
    registry = build_non_strict_registry()
    
    registry.unregister('address')
    registry.register(
        BaseEquals('address'),
        Base32AddressEncoder, CfxAddressDecoder,
        label='address',
    )
    
    return registry



from typing import (
    Set,
    cast,
)

from web3.middleware.cache import (
    construct_simple_cache_middleware
)
from conflux_web3.types import RPCEndpoint

CONFLUX_SIMPLE_CACHE_RPC_WHITELIST = cast(
    Set[RPCEndpoint],
    {
        # "cfx_chainId", this is not a supported api
        # "cfx_gasPrice"
    },
)

simple_cache_middleware = construct_simple_cache_middleware(
    rpc_whitelist=CONFLUX_SIMPLE_CACHE_RPC_WHITELIST,
)



from copy import deepcopy
from typing import (
    Type,
    cast,
    Optional,
    Sequence,
    Tuple,
    TYPE_CHECKING,
    Union,
)
from typing_extensions import (
    Self
)

from hexbytes import (
    HexBytes
)

from ens import (
    ENS,
)
from ens.utils import (
    default,
    normal_name_to_hash,
    raw_name_to_hash,
    label_to_hash,
    normalize_name,
    is_none_or_zero_address, # already hooked
)
from ens.exceptions import (
    UnauthorizedError,
    ResolverNotFound,
)

from cfx_address import (
    Base32Address,
)
from cfx_address.utils import (
    normalize_to
)
from cns.exceptions import (
    InterfaceNotSupported,
    MissingTransactionSender,
    UnstableAPI,
)
from cns.utils import (
    init_web3
)

if TYPE_CHECKING:
    from conflux_web3 import Web3  # noqa: F401
    from conflux_web3.contract import (  # noqa: F401
        ConfluxContract,
    )
    from web3.providers.base import (  # noqa: F401
        BaseProvider,
    )
    from conflux_web3.types import (  # noqa: F401
        Middleware,
        TxParam,
        AddressParam,
    )
    from conflux_web3.types.transaction_hash import (
        TransactionHash
    )

class CNS(ENS):
    w3: "Web3"
    ens: "ConfluxContract"
    allow_unstable_api: bool = False
    _resolver_contract: Type["ConfluxContract"]

    def __init__(
        self,
        provider: "BaseProvider" = cast("BaseProvider", default),
        addr: Optional[Union[Base32Address, str]]=None,
        middlewares: Optional[Sequence[Tuple["Middleware", str]]] = None,
        default_account: Optional[Base32Address] = None,
    ) -> None:
        """
        :param provider: a single provider used to connect to Ethereum
        :type provider: instance of `web3.providers.base.BaseProvider`
        :param hex-string addr: the address of the ENS registry on-chain.
            If not provided, ENS.py will default to the mainnet ENS
            registry address.
        """
        self.w3 = init_web3(provider, middlewares, default_account)

        # if addr is None, propriate address will be automatically selected
        # see conflux_web3.contract.metadata.DEPLOYMENT_INFO for more infomation
        if addr:
            self.ens = self.w3.cfx.contract(addr, name="ENS")
        else:
            self.ens = self.w3.cfx.contract(name="ENS", with_deployment_info=True)
        
        self._resolver_contract = self.w3.cfx.contract(name="RESOLVER", with_deployment_info=False)
        self._reverse_resolver_contract = self.w3.cfx.contract(name="REVERSE_RESOLVER", with_deployment_info=False)
        self._name_wrapper_contract = self.w3.cfx.contract(name="NameWrapper", with_deployment_info=False)
    
    def address(self, name: str) -> Union[Base32Address, None]:
        address = self._resolve(name, "addr")
        return cast(Base32Address, address)
    
    def owner(self, name: str, wrapped: bool=False) -> Base32Address:
        """
        returns the owner of the name.
        pass wrapped = True to know the owner of the name if the the name is wrapped by a NameWrapper

        Parameters
        ----------
        name : str
            the domain name
        wrapped : bool, optional
            whether further get the name from name wrapper, by default False

        Returns
        -------
        Base32Address
            the owner of the name
        """        
        owner = cast(Base32Address, super().owner(name))
        if not wrapped:
            return owner
        return self._owner_from_name_wrapper(owner, name)
    
    def resolver(self, name: str) -> Optional["ConfluxContract"]:
        return cast("ConfluxContract", super().resolver(name))
    
    def setup_owner(
        self,
        name: str,
        new_owner: Union[Base32Address, str] = cast(Base32Address, default),
        transact: Optional["TxParam"] = None,
        wrapped: bool = False
    ) -> Optional[Base32Address]:
        self._check_unstable_api("setup_owner")
        name = normalize_name(name)

        transact = deepcopy(transact) if transact else {}
        (super_owner, unowned, owned) = self._first_owner(name)
        if new_owner == default:
            new_owner = self._tx_sender(transact)
        elif not new_owner:
            new_owner = Base32Address.zero_address(network_id=self.w3.cfx.chain_id)
        # claim subdomain if needed
        self._claim_ownership(new_owner, unowned, owned, transact=transact, wrapped=wrapped)
        return cast(Base32Address, new_owner)
    
    def setup_address(
        self,
        name: str,
        address: Union[Base32Address, str, None] = cast(Base32Address, default),
        transact: Optional["TxParam"] = None,
        wrapped: bool = False
    ) -> Optional["TransactionHash"]:
        """
        NOTE: This api is not stable, you should set ``w3.cns.allow_unstable_api = True`` to enable this api.
        Set up the name to point to the supplied address.
        Make sure the sender of the transaction has the permission to set the address or the transaction execution will fail.

        Parameters
        ----------
        name : str
            name to setup
        address : Union[Base32Address, str, None], optional
            name will point to this address.
            If ``None``, erase the record. 
            If not specified, name will point to the sender of the transaction
        transact : Optional[TxParam], optional
            Specify the non-default transaction execution information.
            Specifically, ``from`` field to specify the transaction sender.
            Required if ``w3.cfx.default_account`` is not set.
            By default None
        wrapped : bool, optional
            whether the name is wrapped by a name wrapper, by default False

        Returns
        -------
        Optional[TransactionHash]
            The transaction hash object of the transaction to set the address. 
            None if the name was already pointed to the target address

        Raises
        ------
        ResolverNotFound
            the resolver of the address is not found
        """
        self._check_unstable_api("setup_address")
        # simple implementation without subnode setup
        transact = deepcopy(transact) if transact else {}
        name = normalize_name(name)
        # address is not provided, so set "from" as the target address
        # NOTE: difference to web3.py - we do not use owner as address, sender is used as address
        if address == default:
            address = self._tx_sender(transact)
        elif address is None:
            address = Base32Address.zero_address(network_id=self.w3.cfx.chain_id)
        if self.address(name) == address:
            return None
        
        self.setup_owner(name, transact=transact, wrapped=wrapped)

        resolver, current_name = self._get_resolver(name)
        assert resolver is not None
        self._set_resolver(name, resolver.address, transact, wrapped=wrapped)
        
        return resolver.functions.setAddr(normal_name_to_hash(name), address).transact(transact)

    def _get_resolver(self, normal_name: str, fn_name: str = "addr") -> Tuple[Optional["ConfluxContract"], str]:
        return super()._get_resolver(normal_name, fn_name) # type: ignore

    def _set_resolver(self, name: str, resolver_addr: Base32Address, transact: "TxParam", wrapped: bool=False) -> "ConfluxContract":
        # if is_none_or_zero_address(resolver_addr):
        #     resolver_addr = self.address("resolver.eth")
        namehash = raw_name_to_hash(name)
        if self.ens.caller.resolver(namehash) != resolver_addr:
            if wrapped:
                self._name_wrapper_contract(self.owner(name)).functions.setResolver(namehash, resolver_addr).transact(transact).executed()
            else:
                self.ens.functions.setResolver(namehash, resolver_addr).transact(transact).executed()
        return self._resolver_contract(address=resolver_addr)

    def setup_name(self, name: str, address: Optional[Base32Address] = None, transact: Optional["TxParam"] = None) -> HexBytes:
        self._check_unstable_api("setup_name")
        return super().setup_name(name, address, transact) # type: ignore
    
    def set_text(self, name: str, key: str, value: str, transact: Optional["TxParam"] = None) -> HexBytes:
        self._check_unstable_api("set_text")
        return super().set_text(name, key, value, transact) # type: ignore

    def _claim_ownership(
        self,
        owner: "AddressParam",
        unowned: Sequence[str],
        owned: str,
        old_owner: Optional[Base32Address] = None,
        transact: Optional["TxParam"] = None,
        wrapped: bool = False,
    ) -> None:
        if not transact:
            transact = {}
        transact = deepcopy(transact)
        # transact["from"] = old_owner or owner
        if wrapped:
            name_wrapper = self._name_wrapper_contract(
                self.owner(owned)
            )
        for label in reversed(unowned):
            # TODO: manually specify nonce
            if wrapped:
                name_wrapper.functions.setSubnodeOwner( # type: ignore
                    raw_name_to_hash(owned), label, owner, 0, 2**64-1
                ).transact(transact).executed()
            else:
                self.ens.functions.setSubnodeOwner(
                    raw_name_to_hash(owned), label_to_hash(label), owner
                ).transact(transact).executed()
            owned = f"{label}.{owned}"
    
    def _tx_sender(self, transact: "TxParam") -> Base32Address:
        if 'from' in transact:
            return cast(Base32Address, transact['from'])
        elif self.w3.cfx.default_account:
            return self.w3.cfx.default_account
        else:
            raise MissingTransactionSender("Transaction sender is not specified: please set w3.cfx.default_account or transact['from'])")

    def _first_owner(self, name: str) -> Tuple[Optional[Base32Address], Sequence[str], str]:
        return cast(Tuple[Optional[Base32Address], Sequence[str], str], super()._first_owner(name))

    def _owner_from_name_wrapper(self, address: Base32Address, name: str) -> Base32Address:
        name_wrapper = self._name_wrapper_contract(address)
        try:
            return name_wrapper.caller.ownerOf(
                int.from_bytes(raw_name_to_hash(name), "big")
            )
        except:
            # TODO: use supportInterface to judge
            raise InterfaceNotSupported(f"Interface ownerOf is not invoked successfully: please check if {address} is a NameWrapper")

    @classmethod
    def fromWeb3(cls, w3: "Web3", addr: Optional[Base32Address] = None) -> Self:
        return cls.from_web3(w3, addr)

    @classmethod
    def from_web3(cls, w3: "Web3", addr: Optional[Base32Address] = None) -> Self:
        """
        Generate an ENS instance with web3

        :param `web3.Web3` w3: to infer connection information
        :param hex-string addr: the address of the ENS registry on-chain. If not
            provided, defaults to the mainnet ENS registry address.
        """
        provider = w3.manager.provider
        middlewares = w3.middleware_onion.middlewares
        default_account = w3.cfx.default_account
        return cls(cast("BaseProvider", provider), addr=addr, middlewares=middlewares, default_account=default_account)

    def _check_unstable_api(self, api_name):
        if not self.allow_unstable_api:
            raise UnstableAPI(f"API {api_name} is forbidden because it is still unstable. You can enable it by setting w3.cns.allow_unstable_api = True")

    def _assert_control(self, account: Base32Address, name: str, parent_owned: Optional[str] = None) -> None:
        # if is contract, check if it is a name wrapper
        # if account.address_type == 'contract':
        #     try:
        #         # we assume 
        #         account = self._owner_from_name_wrapper(account, name)
        #     except InterfaceNotSupported:
        #         raise UnauthorizedError(
        #             f"in order to modify {name!r}, you must control account"
        #             f" {account!r}, which is a contract"
        #         )
        accounts_in_control: Sequence[Base32Address] = [normalize_to(item, self.w3.cfx.chain_id) for item in self.w3.wallet.accounts]
        try:
            accounts_in_control += list(self.w3.cfx.accounts)
        except ValueError as e:
            if e.args[0]['code'] == 32601 and e.args[0]['message'] == 'the method accounts does not exist/is not available':
                pass
        
        for item in accounts_in_control:
            if account == item:
                return # account
        raise UnauthorizedError(
            f"in order to modify {name!r}, you must control account"
            f" {account!r}, which owns {parent_owned or name!r}"
        )




# WARNING: 
# note that the hook won't effect already imported objects,
# so make sure this module is executed before any module of conflux_web3 or web3 is executed
# or the hook might not take effect
# else try directly importing the package to hook

from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Union,
)
from cfx_utils.post_import_hook import (
    when_imported
)
from cfx_address import (
    Base32Address
)
from cfx_utils.exceptions import (
    InvalidEpochNumebrParam
)

if TYPE_CHECKING:
    from conflux_web3 import (
        Web3
    )
    from conflux_web3.types import (
        EpochNumberParam,
    )
    
    

def conditional_func(target_func: Callable, condition: Callable[..., bool]) -> Callable:
    """decorate a function to optionally execute another one
    if condition:
        target_func
    else:
        original_func

    Args:
        target_func (Callable): function to be executed if condition
        condition (Callable[..., bool]): receives func arguments, returns a bool
    """
    def inner(func) :
        def wrapper(*args, **kwargs):
            if condition(*args, **kwargs):
                return target_func(*args, **kwargs)
            return func(*args, **kwargs)
        return wrapper
    return inner

def conditional_post_func(post_func: Callable, condition: Callable[..., bool]) -> Callable:
    """decorate a function to optionally execute post operation
    rtn = original_func
    if condition:
        post_func
    return rtn

    Args:
        post_func (Callable): function to be executed after original function if condition
        condition (Callable[..., bool]): receives func arguments, returns a bool
    """
    def inner(func) :
        def wrapper(*args, **kwargs):
            rtn = func(*args, **kwargs)
            if condition(*args, **kwargs):
                post_func(*args, **kwargs)
            else:
                pass
            return rtn
        return wrapper
    return inner

def cfx_web3_condition(*args, **kwargs) -> bool:
    """

    Returns:
        bool: returns if conflux_web3.Web3 type variable in arguments
    """
    from conflux_web3 import Web3
    for arg in list(args)+list(kwargs.values()):
        if isinstance(arg, Web3):
            return True
    return False

@when_imported("web3._utils.contracts")
def hook_encode_abi(mod):
    if mod.__name__ == "web3._utils.contracts":
        from conflux_web3._utils.contracts import cfx_encode_abi
        mod.encode_abi = conditional_func(
            cfx_encode_abi,
            cfx_web3_condition
        )(mod.encode_abi)

eth_to_cfx_tag_mapping: Dict[str, "EpochNumberParam"] = {
    "pending": "pending",
    "latest": "latest_state",
    "safe": "latest_confirmed",
    "finalized": "latest_finalized",
}

# used to hook web3.contract.parse_block_identifier
# hook is activated in _web3_hook
def cfx_parse_block_identifier(
    w3: "Web3", block_identifier: Union["EpochNumberParam", None]
) -> "EpochNumberParam":
    if block_identifier is None:
        return w3.cfx.default_block
    if isinstance(block_identifier, int):
        return block_identifier
    elif block_identifier in ['earliest', 'latest_checkpoint', 'latest_finalized', 'latest_confirmed', 'latest_state', 'latest_mined']: # get_args("EpochLiteral")
        return block_identifier
    elif isinstance(block_identifier, bytes) or (
        isinstance(block_identifier, str) and block_identifier.startswith("0x")
    ):
        # r = 
        # assert r is not None
        return w3.cfx.get_block_by_hash(block_identifier)["epochNumber"] # type: ignore
    elif block_identifier in eth_to_cfx_tag_mapping:
        return eth_to_cfx_tag_mapping[block_identifier]
    else:
        raise InvalidEpochNumebrParam

@when_imported("web3._utils.contracts")
def hook_parse_block_identifier(mod):
    # from conflux_web3.contract import cfx_parse_block_identifier
    if mod.__name__ == "web3._utils.contracts":
        mod.parse_block_identifier = conditional_func(
            cfx_parse_block_identifier,
            cfx_web3_condition
        )(mod.parse_block_identifier)


# hooked to is_none_or_zero_address in _web3_hook
def is_none_or_base_32_zero_address(addr) -> bool:
    EMPTY_ADDR_HEX = "0x" + "00" * 20
    try:
        return (not addr) or (addr == EMPTY_ADDR_HEX) or (Base32Address.decode(addr)["hex_address"] == EMPTY_ADDR_HEX)
    except:
        return False

@when_imported("ens.utils")
def hook_is_none_or_zero_address(mod):
    if mod.__name__ == "ens.utils":
        mod.is_none_or_zero_address = is_none_or_base_32_zero_address

# we manually import ens.utils to hook is_none_or_zero_address because of insuccessful hook
import ens.utils
